import operator as op

assert op.le(1, 2) == True
assert op.lt(1, 2) == True
assert op.ge(1, 2) == False
assert op.gt(1, 2) == False
assert op.eq(1, 2) == False
assert op.ne(1, 2) == True
assert op.and_(0b01, 0b11) == 0b01
assert op.or_(0b01, 0b11) == 0b11
assert op.xor(0b01, 0b11) == 0b10
assert op.invert(0b01) == -0b10
assert op.lshift(0b01, 1) == 0b10
assert op.rshift(0b10, 1) == 0b01
assert op.is_('a', 'a') == True
assert op.is_not('a', 1) == True
assert op.not_(True) == False
assert op.neg(1) == -1
assert op.truth(1) == True
assert op.contains([1, 2], 1) == True
assert op.add(1, 2) == 3
assert op.sub(1, 2) == -1
assert op.mul(1, 2) == 2
assert op.truediv(1, 2) == 0.5
assert op.floordiv(1, 2) == 0
assert op.mod(1, 2) == 1
assert op.pow(2, 3) == 8

class A:
    def __matmul__(self, other):
        return 'matmul'
assert op.matmul(A(), 1) == 'matmul'

a = [1, 2]
assert op.getitem(a, 0) == 1
op.setitem(a, 0, 3)
assert a == [3, 2]
op.delitem(a, 0)
assert a == [2]

a = 'abc'
assert op.iadd(a, 'def') == 'abcdef'
assert op.isub(8, 3) == 5
assert op.imul(a, 2) == 'abcabc'
assert op.itruediv(8, 2) == 4.0
assert op.ifloordiv(8, 3) == 2
assert op.imod(8, 3) == 2
assert op.iand(0b01, 0b11) == 0b01
assert op.ior(0b01, 0b11) == 0b11
assert op.ixor(0b01, 0b11) == 0b10
assert op.ilshift(0b01, 1) == 0b10
assert op.irshift(0b10, 1) == 0b01


# https://github.com/pocketpy/pocketpy/issues/455
class Person:
    def __init__(self, name, age):
        self.name = name
        self.age = age

person = Person("Alice", 30)
res = op.attrgetter('name')(person)
assert res == "Alice"



a = [1, 2, 3]
a = iter(a)

total = 0

while True:
    try:
        obj = next(a)
    except StopIteration:
        break
    total += obj

assert total == 6

class Task:
    def __init__(self, n):
        self.n = n

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if self.i == self.n:
            raise StopIteration
        self.i += 1
        return self.i

a = Task(3)
assert sum(a) == 6

i = iter(Task(5))
assert next(i) == 1
assert next(i) == 2
assert next(i) == 3
assert next(i) == 4
assert next(i) == 5
try:
    next(i)
    exit(1)
except StopIteration:
    pass

a = iter([1])
assert next(a) == 1

try:
    next(a)
    exit(1)
except StopIteration:
    pass




assert type(1) is int
assert type(1.0) is float
assert type(object) is type
assert type(type) is type

assert hasattr(object, '__base__')
assert hasattr(1, '__add__')
assert hasattr(int, '__add__')

assert type(1).__add__(1, 2) == 3
assert getattr(1, '__add__')(2) == 3

a = object()
setattr(a, 'b', 1)
assert a.b == 1
assert hasattr(a, 'b')
assert getattr(a, 'b') == 1
assert getattr(a, 'c', ...) == ...
delattr(a, 'b')
assert not hasattr(a, 'b')

try:
    getattr(a, 'xxx')
    exit(1)
except AttributeError:
    pass

assert getattr(a, 'xxx', 1) == 1

class A:
    def __init__(self, x):
        self.x = x

    def __getattr__(self, name):
        if not name:
            raise AttributeError
        return name, None
    
a = A(1)
assert a.x == 1
assert a.y == ('y', None)
assert a.zzz == ('zzz', None)

assert getattr(a, 'x') == 1
assert getattr(a, 'zzz') == ('zzz', None)

assert hasattr(a, 'x')
assert hasattr(a, 'y')
assert hasattr(a, 'zzz')

assert not hasattr(a, '')




import os
import sys
import shutil

assert os.system("python prebuild.py") == 0

if not os.path.exists("build"):
    os.mkdir("build")

# python cmake_build.py [Debug|Release|RelWithDebInfo] ...

if len(sys.argv) > 1:
    config = sys.argv[1]
else:
    config = 'Release'

extra_flags = " ".join(sys.argv[2:])

assert config in ['Debug', 'Release', 'RelWithDebInfo']

os.chdir("build")

code = os.system(f"cmake .. -DPK_ENABLE_MIMALLOC=ON -DCMAKE_BUILD_TYPE={config} {extra_flags}")
assert code == 0
code = os.system(f"cmake --build . --config {config} -j 4")
assert code == 0

if sys.platform == "win32":
    shutil.copy(f"{config}/main.exe", "../main.exe")
    dll_path = f"{config}/pocketpy.dll"
    if os.path.exists(dll_path):
        shutil.copy(dll_path, "../pocketpy.dll")
elif sys.platform == "darwin":
    shutil.copy("main", "../main")
    dll_path = "libpocketpy.dylib"
    if os.path.exists(dll_path):
        shutil.copy(dll_path, "../libpocketpy.dylib")
else:
    shutil.copy("main", "../main")
    dll_path = "libpocketpy.so"
    if os.path.exists(dll_path):
        shutil.copy(dll_path, "../libpocketpy.so")



from typing import Self, ClassVar

intptr = int

def malloc(size: int) -> intptr: ...
def free(ptr: intptr) -> None: ...

def memcpy(dst: intptr, src: intptr, n: int) -> None: ...
def memset(s: intptr, c: int, n: int) -> None: ...
def memcmp(s1: intptr, s2: intptr, n: int) -> int: ...

def read_cstr(p: intptr) -> str: ...
def read_bytes(p: intptr, n: int) -> bytes: ...
def write_cstr(p: intptr, data: str) -> None: ...
def write_bytes(p: intptr, data: bytes) -> None: ...

class Memory:
    size: ClassVar[int]

class _BuiltinMemory[T](Memory):
    value: T

    def __new__(cls, value: T | None = None) -> None: ...
    @staticmethod
    def read(p: intptr, offset: int) -> T: ...
    @staticmethod
    def write(p: intptr, offset: int, value: T) -> None: ...
    @classmethod
    def array(cls, length: int) -> Self: ...

    def __getitem__(self, index: int) -> T: ...
    def __setitem__(self, index: int, value: T) -> None: ...

class Char(_BuiltinMemory[int]): ...
class UChar(_BuiltinMemory[int]): ...
class Short(_BuiltinMemory[int]): ...
class UShort(_BuiltinMemory[int]): ...
class Int(_BuiltinMemory[int]): ...
class UInt(_BuiltinMemory[int]): ...
class Long(_BuiltinMemory[int]): ...
class ULong(_BuiltinMemory[int]): ...
class LongLong(_BuiltinMemory[int]): ...
class ULongLong(_BuiltinMemory[int]): ...

class Float(_BuiltinMemory[float]): ...
class Double(_BuiltinMemory[float]): ...
class Pointer(_BuiltinMemory[intptr]): ...
class Bool(_BuiltinMemory[bool]): ...

Int8: _BuiltinMemory[int]
UInt8: _BuiltinMemory[int]
Int16: _BuiltinMemory[int]
UInt16: _BuiltinMemory[int]
Int32: _BuiltinMemory[int]
UInt32: _BuiltinMemory[int]
Int64: _BuiltinMemory[int]
UInt64: _BuiltinMemory[int]
SizeT: _BuiltinMemory[int]

IntptrT: _BuiltinMemory[int]
UintptrT: _BuiltinMemory[int]

def addressof(obj: Memory) -> intptr: ...
def sizeof(obj: type[Memory]) -> int: ...




try:
    import msgpack
except ImportError:
    print('msgpack is not enabled, skipping test...')
    exit()

a = {
    'a': 1,
    'b': 2,
    'c': None,
    'd': [1, 2, 3],
    'e': {
        'a': 100,
        'b': 2.5,
        'c': None,
        'd': [142, 2785, 39767],
    },
    "f": 'This is a string',
    'g': [True, False, None],
    'h': False
}

import msgpack

assert msgpack.loads(b'\x01') == 1
assert msgpack.loads(b'\xa11') == "1"
assert msgpack.loads(b'\xcb\x00\x00\x00\x00\x00\x00\x00\x00') == 0.0
assert msgpack.loads(b'\x92\x01\x02') == [1, 2]
assert msgpack.loads(b'\xc0') == None
assert msgpack.loads(b'\xc3') == True
assert msgpack.loads(b'\xc2') == False
assert msgpack.loads(b'\x80') == {}

_j = msgpack.dumps(a)
_a = msgpack.loads(_j)

for k, v in a.items():
    assert (a[k] == _a[k]), f'{a[k]} != {_a[k]}'

for k, v in _a.items():
    assert (a[k] == _a[k]), f'{a[k]} != {_a[k]}'

b = [1, 2, True, None, False]

_j = msgpack.dumps(b)
_b = msgpack.loads(_j)

assert b == _b

c = 1.0
_j = msgpack.dumps(c)
_c = msgpack.loads(_j)
assert c == _c

d = True
_j = msgpack.dumps(d)
_d = msgpack.loads(_j)
assert d == _d

# assert msgpack.dumps((1,)) == '[1]'
# assert msgpack.dumps((1, 2, 3)) == '[1, 2, 3]'
# assert msgpack.dumps(tuple()) == '[]'

assert msgpack.dumps([]) == b'\x90'
assert msgpack.dumps([1, 2, 3]) == b'\x93\x01\x02\x03'
assert msgpack.dumps([1]) == b'\x91\x01'

try:
    msgpack.dumps({1: 2})
    assert False
except TypeError:
    assert True

try:
    msgpack.dumps(type)
    assert False
except TypeError:
    assert True



import pickle

data1 = [1, 2, 3] * 100
data2 = [1.0, 2.0, 3.0] * 100
data3 = ['abcdefg', 'hijklmn', '_______________1'] * 100
data4 = [True, False, True] * 100
data5 = [None, None] * 100

original = {
    '1': data1,
    '2': data2,
    '3': data3,
    '45': {
        '4': data4,
        '5': data5,
    }
}

for i in range(10000):
    encoded = pickle.dumps(original)
    decoded = pickle.loads(encoded)
    if i == 0:
        assert original == decoded




# only one level nested closure is implemented

def f0(a, b):
    def f1():
        return a + b
    return f1

a = f0(1, 2)
b = f0(3, 4)
assert a() == 3
assert b() == 7


def f0(a, b):
    def f1():
        a = 5   # use this first
        return a + b
    return f1

a = f0(1, 2)
assert a() == 7

def f3(x, y):
    return lambda z: x + y + z

a = f3(1, 2)
assert a(3) == 6

# closure ex
def f(n):
    def g(x):
        if x==n:
            return n
        return g(x+1)
    return g(0)

assert f(10) == 10

# class closure
class A:
    def g(self, x):
        def f(y):
            return x + y
        return f
    
assert A().g(1)(2) == 3

# closure with yield
def g(x):
    def fx(y):
        yield x
        yield y
        return x + y
    return fx
    
gen = g(1)(2)
assert next(gen) == 1
assert next(gen) == 2
try:
    next(gen)
    assert False
except StopIteration as e:
    assert e.value == 3



# https://docs.python.org/3/library/operator.html#mapping-operators-to-functions

def le(a, b): return a <= b
def lt(a, b): return a < b
def ge(a, b): return a >= b
def gt(a, b): return a > b
def eq(a, b): return a == b
def ne(a, b): return a != b

def and_(a, b): return a & b
def or_(a, b): return a | b
def xor(a, b): return a ^ b
def invert(a): return ~a
def lshift(a, b): return a << b
def rshift(a, b): return a >> b

def is_(a, b): return a is b
def is_not(a, b): return a is not b
def not_(a): return not a
def truth(a): return bool(a)
def contains(a, b): return b in a

def add(a, b): return a + b
def sub(a, b): return a - b
def mul(a, b): return a * b
def truediv(a, b): return a / b
def floordiv(a, b): return a // b
def mod(a, b): return a % b
def pow(a, b): return a ** b
def neg(a): return -a
def matmul(a, b): return a @ b

def getitem(a, b): return a[b]
def setitem(a, b, c): a[b] = c
def delitem(a, b): del a[b]

def iadd(a, b): a += b; return a
def isub(a, b): a -= b; return a
def imul(a, b): a *= b; return a
def itruediv(a, b): a /= b; return a
def ifloordiv(a, b): a //= b; return a
def imod(a, b): a %= b; return a
# def ipow(a, b): a **= b; return a
# def imatmul(a, b): a @= b; return a
def iand(a, b): a &= b; return a
def ior(a, b): a |= b; return a
def ixor(a, b): a ^= b; return a
def ilshift(a, b): a <<= b; return a
def irshift(a, b): a >>= b; return a

class attrgetter:
    def __init__(self, attr):
        self.attr = attr
    def __call__(self, obj):
        return getattr(obj, self.attr)



from .writer import Writer
from .schema import Enum

def gen_enum(w: Writer, pyi_w: Writer, enum: Enum):
    for value in enum.values:
        w.write(f'ADD_ENUM({value.name});')

        if value.value is not None:
            pyi_w.write(f'{value.name}: int = {value.value}')
        else:
            pyi_w.write(f'{value.name}: int')
        


from enum import Enum

class A(Enum):
    a = 1
    b = '2'
    c = None

assert str(A.a) == 'A.a'
assert repr(A.a) == '<A.a: 1>'

assert str(A.b) == 'A.b'
assert repr(A.b) == "<A.b: '2'>"

assert str(A.c) == 'A.c'
assert repr(A.c) == '<A.c: None>'

assert A.a == A.a
assert A.a != A.b
assert A.a != A.c

assert A.a.name == 'a'
assert A.a.value == 1

assert A.b.name == 'b'
assert A.b.value == '2'

assert A.c.name == 'c'
assert A.c.value is None



try:
    import os
except ImportError:
    print('os is not enabled, skipping test...')
    exit(0)

import sys
if sys.platform == 'win32':
    exe_name = 'main.exe'
else:
    exe_name = './main'
assert os.system(f'{exe_name} --compile python/heapq.py heapq1.pyc') == 0
assert os.path.exists('heapq1.pyc')

import heapq1
import heapq

a = [1, 2, -3, 2, 1, 5, 11, 123] * 10
b = a.copy()

heapq.heapify(a)
heapq1.heapify(b)

assert a == b

os.remove('heapq1.pyc')



# test == !=
assert True == True
assert True != False
assert False == False
assert False != True

# test and/or/not
assert True and True
assert not (True and False)
assert True or True
assert True or False
assert not False
assert not (not True)

assert bool() == False
assert bool(0) == False
assert bool(1) == True
assert bool([]) == False
assert bool("abc") == True
assert bool([1,2]) == True
assert bool('') == False

# is operator
assert None == None
assert None is None
assert ... == ...
assert ... is ...
assert NotImplemented == NotImplemented
assert NotImplemented is NotImplemented

assert True is True
assert False is False

assert True + 1 == 2
assert True - 1 == 0
assert True * 3 == 3
assert 1 + True == 2
assert 1 - True == 0
assert 3 * True == 3



from typing import Self, Literal
from vmath import vec2, vec2i

class TValue[T]:
    def __new__(cls, value: T) -> Self: ...
    
    @property
    def value(self) -> T: ...

# TValue_int = TValue[int]
# TValue_float = TValue[float]
# TValue_vec2i = TValue[vec2i]
# TValue_vec2 = TValue[vec2]
    
configmacros: dict[str, int]

def memory_usage_info() -> str: ...
def memory_usage() -> int: ...


def currentvm() -> int:
    """Return the current VM index."""


def watchdog_begin(timeout: int):
    """Begin the watchdog with `timeout` in milliseconds.

    `PK_ENABLE_WATCHDOG` must be defined to `1` to use this feature.
    You need to call `watchdog_end()` later.
    If `timeout` is reached, `TimeoutError` will be raised.
    """
def watchdog_end() -> None:
    """End the watchdog after a call to `watchdog_begin()`."""

def profiler_begin() -> None: ...
def profiler_end() -> None: ...
def profiler_reset() -> None: ...
def profiler_report() -> dict[str, list[list]]: ...

class ComputeThread:
    def __init__(self, vm_index: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]): ...

    @property
    def is_done(self) -> bool:
        """Check if the current job is done."""

    def wait_for_done(self) -> None:
        """Wait for the current job to finish."""

    def last_error(self) -> str | None: ...
    def last_retval(self): ...

    def submit_exec(self, source: str) -> None:
        """Submit a job to execute some source code."""

    def submit_eval(self, source: str) -> None:
        """Submit a job to evaluate some source code."""

    def submit_call(self, eval_src: str, *args, **kwargs) -> None:
        """Submit a job to call a function with arguments."""

    def exec(self, source: str) -> None:
        """Directly execute some source code."""

    def eval(self, source: str):
        """Directly evaluate some source code."""



import easing

def validate(k, f, x):
    val = f(x)
    assert (-2 <= val <= 2.0), (k, x, val)
    assert isinstance(val, float), (k, x, val)

for k, f in easing.__dict__.items():
    if callable(f):
        validate(k, f, 0.2)
        validate(k, f, 0.5)
        validate(k, f, 0.8)
        validate(k, f, 1.0)



import time

def costly_func(n: int):
    time.sleep(n)               # 2s
    time.sleep(1)               # 1s

x = 1
y = 2
costly_func(2)      # 3s

time.sleep(1)       # 1s

def new_func(a, b, c):
    x = a + b
    x = x + c
    return a

def new_func2(a, b, c):
    x = a + b
    x = x + c
    return a



from typing import Literal
from .schema import *

class UnsupportedNode(Exception):
    def __init__(self, node: c_ast.Node):
        self.node = node

    def __str__(self):
        return f"'{type(self.node)}' is not supported\n{self.node!r}"
    

class Header:
    def __init__(self):
        self.types = [] # type: list
        self.type_aliases = {} # type: dict[str, str | c_ast.Node]
        self.builtin_aliases = {
            'size_t', 'bool',
            'int8_t', 'int16_t', 'int32_t', 'int64_t',
            'uint8_t', 'uint16_t', 'uint32_t', 'uint64_t',
            'intptr_t', 'uintptr_t',
        }
        self.functions = [] # type: list[Function]

    def remove_types(self, names: set):
        self.types = [t for t in self.types if getattr(t, 'name', None) not in names]

    def remove_functions(self, names: set):
        self.functions = [f for f in self.functions if f.name not in names]

    def add_type_alias(self, k: str, v: str | c_ast.Node):
        if k in self.builtin_aliases:
            return
        self.type_aliases[k] = v

    def build_enum(self, node: c_ast.Enum, alias_name: str | None = None):
        enum = Enum(node.name, alias_name)
        for item in node.values.enumerators:
            enum.values.append(item.name)
        self.types.append(enum)

    def build_struct(self, node: c_ast.Struct | c_ast.Union, alias_name: str | None = None):
        if isinstance(node, c_ast.Struct):
            cls = Struct
        elif isinstance(node, c_ast.Union):
            cls = Union
        else:
            raise UnsupportedNode(node)
        if node.decls is not None:
            fields = {}
            for decl in node.decls:
                try:
                    type, name = self.build_param(decl)
                    assert name
                    fields[name] = type
                except UnsupportedNode:
                    pass
            self.types.append(cls(node.name, fields, alias_name))
        else:
            self.types.append(cls(node.name, None, alias_name))

    def build_type(self, node, alias_name):
        if isinstance(node, c_ast.Enum):
            self.build_enum(node, alias_name)
            if alias_name:
                self.add_type_alias(alias_name, node)
        elif isinstance(node, (c_ast.Struct, c_ast.Union)):
            self.build_struct(node, alias_name)
            if alias_name:
                self.add_type_alias(alias_name, node)
        elif isinstance(node, c_ast.IdentifierType):
            assert alias_name
            assert node.names[0]
            self.add_type_alias(alias_name, node.names[0])
        else:
            raise UnsupportedNode(node)
        
    def get_type_name(self, node):
        # convert array to const T*
        if isinstance(node, c_ast.ArrayDecl):
            dims = []
            while isinstance(node, c_ast.ArrayDecl):
                dims.append(node.dim.value)
                node = node.type
            base_name = self.get_type_name(node)
            # return base_name + ''.join(f'[{dim}]' for dim in dims)
            return f'const {base_name}*'

        node, level = self.eat_pointers(node)

        if isinstance(node, c_ast.FuncDecl):
            assert level == 1
            return 'void (*)()'

        if not isinstance(node, (c_ast.TypeDecl, c_ast.Typename)):
            raise UnsupportedNode(node)

        is_const = node.quals and 'const' in node.quals
        node = node.type
        base_name: str
        if isinstance(node, c_ast.IdentifierType):
            base_name = node.names[0]
        elif isinstance(node, c_ast.Enum):
            base_name = 'enum ' + node.name
        elif isinstance(node, c_ast.Struct):
            base_name = 'struct ' + node.name
        elif isinstance(node, c_ast.Union):
            base_name = 'union ' + node.name
        else:
            base_name = self.get_type_name(node)

        if is_const:
            base_name = 'const ' + base_name
        return base_name + '*' * level
        
    def build_param(self, node):
        name = None
        if isinstance(node, c_ast.Decl):
            name = node.name
            node = node.type
        return self.get_type_name(node), name
    
    def eat_pointers(self, node):
        level = 0
        while isinstance(node, c_ast.PtrDecl):
            node = node.type
            level += 1
        return node, level
        
    def build_function(self, node: c_ast.FuncDecl):
        args = node.args
        node = node.type
        node, level = self.eat_pointers(node)
        assert isinstance(node, c_ast.TypeDecl), type(node)
        name = node.declname
        ret = node.type.names[0]
        is_const = node.quals and 'const' in node.quals
        if is_const:
            ret = 'const ' + ret
        func = Function(name, ret + '*' * level)
        if args is not None:
            for param in args.params:
                if isinstance(param, c_ast.EllipsisParam):
                    func.args.append('...')
                else:
                    T, name = self.build_param(param)
                    if T != 'void':
                        func.args.append((T, name))
        self.functions.append(func)

    def build(self, ast: c_ast.FileAST):
        for _, node in ast.children():
            if isinstance(node, c_ast.Typedef):
                name, node = node.name, node.type
                if isinstance(node, c_ast.TypeDecl):
                    self.build_type(node.type, name)
                elif isinstance(node, c_ast.PtrDecl):
                    self.add_type_alias(name, node)
                else:
                    raise UnsupportedNode(node.type)
            elif isinstance(node, c_ast.Decl):
                if isinstance(node.type, c_ast.FuncDecl):
                    self.build_function(node.type)
                else:
                    try:
                        self.build_type(node.type, None)
                    except UnsupportedNode:
                        pass
            else:
                raise UnsupportedNode(node.type)


try:
    import os
    import io
except ImportError:
    print('os is not enabled, skipping test...')
    exit(0)

a = open('123.txt', 'wt')
a.write('123')
a.write('456')
a.close()

with open('123.txt', 'rt') as f:
    assert f.read() == '123456'

with open('123.txt', 'rt') as f:
    assert f.read(3) == '123'
    assert f.tell() == 3
    assert f.read(3) == '456'
    assert f.tell() == 6
    assert f.read(3) == ''      # EOF
    assert f.tell() == 6

with open('123.txt', 'rb') as f:
    assert f.read(2) == b'12'
    assert f.tell() == 2
    assert f.read(2) == b'34'
    assert f.tell() == 4
    assert f.read(2) == b'56'
    assert f.tell() == 6
    assert f.read(2) == b''     # EOF
    assert f.tell() == 6

# test fseek
with open('123.txt', 'rt') as f:
    f.seek(0, io.SEEK_END)
    assert f.tell() == 6
    assert f.read() == ''
    f.seek(3, io.SEEK_SET)
    assert f.tell() == 3
    assert f.read() == '456'
    assert f.tell() == 6

with open('123.txt', 'a') as f:
    f.write('测试')

with open('123.txt', 'r') as f:
    assert f.read() == '123456' + '测试'

assert os.path.exists('123.txt')
os.remove('123.txt')
assert not os.path.exists('123.txt')


with open('123.bin', 'wb') as f:
    f.write('123'.encode())
    f.write('测试'.encode())

def f_():
    with open('123.bin', 'rb') as f:
        b = f.read()
        assert isinstance(b, bytes)
        assert b == '123测试'.encode()

f_()

assert os.path.exists('123.bin')
os.remove('123.bin')
assert not os.path.exists('123.bin')


import math

def generate_sin_table(num_points = 4096 + 1):
    y_step = 1.0 / (num_points - 1)
    sin_table = []
    for i in range(num_points):
        y = i * y_step
        x = math.asin(y)
        x = max(0, min(x, math.pi / 2))
        sin_table.append(x)
    return sin_table

if __name__ == "__main__":
    sin_table = generate_sin_table()
    print(len(sin_table))
    # print(sin_table)

    filepath = 'include/pocketpy/common/_sin_table.h'
    with open(filepath, 'w') as f:
        f.write('static const double _sin_table[] = {')
        for i, val in enumerate(sin_table):
            if i % 8 == 0:
                f.write('\n    ')
            f.write(str(val) + ', ')
        f.write('\n};\n')


from typing import Callable, Literal

def isenabled() -> bool:
    """Check if automatic garbage collection is enabled."""

def enable() -> None:
    """Enable automatic garbage collection."""

def disable() -> None:
    """Disable automatic garbage collection."""

def collect() -> int:
    """Run a full collection immediately.

    Returns an integer indicating the number of unreachable objects found.
    """

def collect_hint() -> int:
    """Hint the garbage collector to run a collection.

    The typical usage scenario for this function is in frame-driven games,
    where `gc.disable()` is called at the start of the game,
    and `gc.collect_hint()` is called at the end of each frame.
    """

def setup_debug_callback(cb: Callable[[Literal['start', 'stop'], str], None] | None) -> None:
    """Setup a callback that will be triggered at the end of each collection."""



import sys
import os

if len(sys.argv) != 4:
    print('Usage: python compileall.py <pocketpy_executable> <source_dir> <output_dir>')
    exit(1)

pkpy_exe = sys.argv[1]
source_dir = sys.argv[2]
output_dir = sys.argv[3]

def do_compile(src_path, dst_path):
    assert os.path.isfile(src_path)
    cmd = f'{pkpy_exe} --compile "{src_path}" "{dst_path}"'
    if os.system(cmd) != 0:
        print(src_path)
        exit(1)

if os.path.isfile(source_dir):
    if output_dir.endswith('.py'):
        output_dir += 'c'
    do_compile(source_dir, output_dir)
    exit(0)

for root, _, files in os.walk(source_dir):
    for file in files:
        if not file.endswith('.py'):
            continue
        src_path = os.path.join(root, file)
        dst_path = os.path.join(
            output_dir,
            os.path.relpath(root, source_dir),
            file + 'c'
        )
        os.makedirs(os.path.dirname(dst_path), exist_ok=True)
        do_compile(src_path, dst_path)



try:
    import os
except ImportError:
    exit(0)

import sys
is_pyc = sys.argv[0].endswith('.pyc')

if is_pyc:
    os.chdir('tmp/tests')
else:
    os.chdir('tests')

assert os.getcwd().endswith('tests')

os.environ['STDOUT'] = ''

import test.a.g.q

assert os.environ['STDOUT'] == 'test init!!\ntest.a init!!\ntest.a.g init!!\ntest.a.g.q init!!\n'

import test

assert test.__package__ == 'test'
assert test.__name__ == 'test'



try:
    from line_profiler import LineProfiler
    print('[INFO] line_profiler is used')
except ImportError:
    exit(0)

def f2(x):
    a = 0
    for i in range(x):
        if i % 5 == 0:
            a += i
    return a

def f1(x):
    res = f2(x)
    return res

lp = LineProfiler()

lp.add_function(f2)

# lp.runcall(f2, 1000000)
# lp.print_stats()
###############################

lp.add_function(f1)
lp.runcall(f1, 1000000)
lp.print_stats()




# https://gist.github.com/dean0x7d/df5ce97e4a1a05be4d56d1378726ff92

a = 1
my_locals = {"b": 2}

# With user-defined locals:
exec("""
import sys
assert "sys" in locals()
assert "sys" not in globals()
assert "a" not in locals()
assert "a" in globals()
# print(a)  # checks `locals()` first, fails, but finds it in `globals()`
assert (a == 1), a
assert "b" in locals()
assert "b" not in globals()
# print(b)
assert (b == 2), b
def main():
    assert "sys" not in locals()   # not the same `locals()` as the outer scope
    assert "sys" not in globals()  # and `sys` isn't in `globals()`, same as before
    assert "b" not in locals() # again, not the same `locals()` as the outer scope
main()
""", globals(), my_locals)

assert "sys" in my_locals  # side effect
assert "sys" not in globals()


# With default locals:
exec("""
import sys
assert locals() == {}
assert "sys" in globals()
def main():
    assert "sys" not in locals()  # not the same locals as the outer scope
    assert "sys" in globals()     # but now be can access `sys` via `globals()`
main()
""", globals())

assert "sys" in globals()


import os
import sys
import time
import subprocess

pkpy_exe = 'main.exe' if sys.platform == 'win32' else './main'

def test_file(filepath, cpython=False):
    if cpython:
        return os.system("python " + filepath) == 0
    code = os.system(pkpy_exe + ' ' + filepath)
    if code != 0:
        print('Return code:', code)
    return code == 0

def test_dir(path):
    print("Testing directory:", path)
    for filename in sorted(os.listdir(path)):
        if not filename.endswith('.py') and not filename.endswith('.pyc'):
            continue
        filepath = os.path.join(path, filename)
        print("> " + filepath, flush=True)

        if path == 'benchmarks/':
            _0 = time.time()
            if not test_file(filepath, cpython=True):
                print('cpython run failed')
                continue
            _1 = time.time()
            if not test_file(filepath): exit(1)
            _2 = time.time()
            print(f'  cpython:  {_1 - _0:.6f}s (100%)')
            print(f'  pocketpy: {_2 - _1:.6f}s ({(_2 - _1) / (_1 - _0) * 100:.2f}%)')
        else:
            if not test_file(filepath):
                print('-' * 50)
                print("TEST FAILED!")
                exit(1)
                # print("TEST FAILED! Press any key to continue...")
                # input()

print('CPython:', str(sys.version).replace('\n', ''))
print('System:', '64-bit' if sys.maxsize > 2**32 else '32-bit')

def test_repl():
    print("[REPL Test Enabled]")
    if sys.platform in ['linux', 'darwin']:
        cmd = './main'
    else:
        cmd = None

    if cmd is not None:
        res = subprocess.run([cmd], encoding='utf-8', input=r'''
def add(a, b):
    return a + b

class A:
    def __init__(self, x):
        self.x = x
    def get(self):
        return self.x

print('ans_1:', add(1, 2))
print('ans_2:', A('abc').get())
exit()
''', capture_output=True, check=True)
        res.check_returncode()
        # assert 'ans_1: 3' in res.stdout, res.stdout
        if 'ans_1: 3' not in res.stdout:
            print(res.stdout)
            exit(1)
        # assert 'ans_2: abc' in res.stdout, res.stdout
        if 'ans_2: abc' not in res.stdout:
            print(res.stdout)
            exit(1)

code = os.system(f'python compileall.py {pkpy_exe} tests tmp/tests')
assert code == 0

if len(sys.argv) == 2:
    assert 'benchmark' in sys.argv[1]
    test_dir('benchmarks/')
else:
    test_dir('tests/')
    test_repl()

    if os.path.exists('tmp/tests'):
        print('-' * 50)
        time.sleep(3)
        test_dir('tmp/tests/')
    


print("ALL TESTS PASSED")



from array2d import array2d, chunked_array2d
from vmath import color32, vec2i

'''
src/modules/array2d.c
'''

# =====array2d_like_render
data = [
    [1,2,3],
    [4,5,6],
]
assert array2d.fromlist(data).render() == "123\n456"

# ====array2d_like_render_with_color
text_data = []
color_fg = []
color_bg = []
for i in range(51):
    r = i * 5  # r in [0,5,10,15...250,255]
    g = i * 5  # g in [0,5,10,15...250,255]
    
    row_fg = []
    row_bg = []
    row_text_data = []
    for j in range(51):
        b = j * 5  # j in [0,5,10,15...250,255]
        row_fg.append(color32(255-r, 255-g, 255-b, 255))
        row_bg.append(color32(r, g, b, 255))
        row_text_data.append("A")
    
    color_fg.append(row_fg)
    color_bg.append(row_bg)
    text_data.append(row_text_data)


array2d.fromlist(text_data).render_with_color(
    array2d.fromlist(color_fg),
    array2d.fromlist(color_bg)
    )

array2d.fromlist(text_data).render_with_color(
    array2d.fromlist(color_fg),
    array2d(51, 51, None)
    )

array2d.fromlist(text_data).render_with_color(
    array2d(51, 51, None),
    array2d.fromlist(color_bg)
    )

# (curr_fg.u32 != 0 || curr_bg.u32 != 0) == false
array2d.fromlist(text_data).render_with_color(
    array2d(51, 51, None),
    array2d(51, 51, None)
    )

try:
    array2d.fromlist(text_data).render_with_color(
        array2d.fromlist(color_fg),
        array2d(51, 51, 1)
        )
    assert False, "here expect a TypeError"
except TypeError as e:
    assert "color32" in str(e) or "None" in str(e)

try:
    array2d.fromlist(text_data).render_with_color(
        array2d(51, 51, 1),
        array2d.fromlist(color_bg)
        )
    assert False, "here expect a TypeError"
except TypeError as e:
    assert "color32" in str(e) or "None" in str(e)


# ====array2d_like_any
data = [[False, False], [False, False]]
assert array2d.fromlist(data).any() == False

# ====c11_array2d_view__set
data = chunked_array2d(4)
data[vec2i(3,3)] = 0
data.view()[1,1] = 10
assert data.view() == array2d.fromlist([
    [None, None, None, None,],
    [None, 10, None, None,],
    [None, None, 0, None,],
    [None, None, None, None,],
    ])

# ====array2d_view_origin
assert data.view()[vec2i(1,1)-data.view().origin] == data[vec2i(1,1)]
assert data.view()[vec2i(3,3)-data.view().origin] == data[vec2i(3,3)]

# ====chunked_array2d__len__
data = chunked_array2d(4)
for i in range(10):
    for j in range(10):
        data[vec2i(i,j)] = 10

assert len(data) == 9

# ====c11_chunked_array2d__mark
import gc

def gc_collect_callback(statue, msg):
    if statue == 'stop':
        for line in msg.split('\n'):
            print(line)
            if "5290" in line and "str" in line:
                return
        print(msg)
        raise AssertionError("Expected GC output not found")

gc.collect()
gc.setup_debug_callback(gc_collect_callback)
def create_garbage():
    data = chunked_array2d(4)
    # [str  ] small:      0  large:   5290
    for x in range(23):
        for y in range(230):
            data[vec2i(x, y)] = "garbage"*100
            
create_garbage()
gc.collect()


# ====chunked_array2d_view_chunks
data = chunked_array2d(4, default = 0)
data[vec2i(-10,-10)] = -1
assert data.view_chunks(vec2i(-3, -3), 6, 6)[vec2i(2, 2)] == -1
data.view_chunks(vec2i(-3, -3), 6, 6)[vec2i(22, 22)] = 1
assert data[vec2i(10,10)] == 1





import os

def get_sources():
    sources = {}
    for file in sorted(os.listdir("python")):
        if not file.endswith(".py"):
            continue
        key = file.split(".")[0]
        const_char_array = []
        with open("python/" + file) as f:
            specials = { 10: '\\n', 34: '\\"' }
            for c in f.read().encode('utf-8'):
                if c in specials:
                    const_char_array.append(specials[c])
                elif c >= 32 and c <= 126 and c != 92:
                    const_char_array.append(chr(c))
                else:
                    const_char_array.append(f'\\x{c:02x}')
        const_char_array = ''.join(const_char_array)
        sources[key] = '"' + const_char_array + '"'
    return sources

sources = get_sources()

# use LF line endings instead of CRLF
with open("include/pocketpy/common/_generated.h", "wt", encoding='utf-8', newline='\n') as f:
    data = '''#pragma once
// generated by prebuild.py

const char* load_kPythonLib(const char* name);

'''
    for key in sorted(sources.keys()):
        value = sources[key]
        data += f'extern const char kPythonLibs_{key}[];\n'
    f.write(data)

with open("src/common/_generated.c", "wt", encoding='utf-8', newline='\n') as f:
    data = '''// generated by prebuild.py
#include "pocketpy/common/_generated.h"
#include <string.h>
'''
    for key in sorted(sources.keys()):
        value = sources[key]
        data += f'const char kPythonLibs_{key}[] = {value};\n'
    f.write(data)

    f.write("\n")
    f.write("const char* load_kPythonLib(const char* name) {\n")
    f.write("    if (strchr(name, '.') != NULL) return NULL;\n")
    for key in sorted(sources.keys()):
        if key.startswith('_'):
            continue
        f.write(f'    if (strcmp(name, "{key}") == 0) return kPythonLibs_{key};\n')
    f.write("    return NULL;\n")
    f.write("}\n")




import pickle as pkl

def test(data): # type: ignore
    print('-'*50)
    b = pkl.dumps(data)
    print(b)
    o = pkl.loads(b)
    print(o)
    assert data == o
    return o

test(None)                      # PKL_NONE
test(...)                       # PKL_ELLIPSIS
test(1)                         # PKL_INT8
test(277)                       # PKL_INT16
test(-66666)                    # PKL_INT32
test(0xffffffffffff)            # PKL_INT64
test(1.0)                       # PKL_FLOAT32
test(1.12312434234)             # PKL_FLOAT64
test(True)                      # PKL_TRUE
test(False)                     # PKL_FALSE
test("hello")                   # PKL_STRING
test(b"hello")                  # PKL_BYTES

from vmath import vec2, vec3, vec2i, vec3i

test(vec2(2/3, 1.0))            # PKL_VEC2
test(vec3(2/3, 1.0, 3.0))       # PKL_VEC3
test(vec2i(1, 2))               # PKL_VEC2I
test(vec3i(1, 2, 3))            # PKL_VEC3I

test(vec3i)                     # PKL_TYPE

print('-'*50)
from array2d import array2d
a = array2d[int | bool | vec2i].fromlist([
    [1, 2, vec2i.LEFT],
    [4, True, 6]
])
a_encoded = pkl.dumps(a)
print(a_encoded)
a_decoded = pkl.loads(a_encoded)
assert isinstance(a_decoded, array2d)
assert a_decoded.width == 3 and a_decoded.height == 2
assert (a == a_decoded).all()
print(a_decoded)

test([1, 2, 3])                 # PKL_LIST
test((1, 2, 3))                 # PKL_TUPLE
test({1: 2, 3: 4})              # PKL_DICT

# test complex data
test([1, '2', True])
test([1, '2', 3.0, True])
test([1, '2', True, {'key': 4}])
test([1, '2', 3.0, True, {'k1': 4, 'k2': [b'xxxx']}])

# test memo
a = [1, 2, 3, 4, 5, 6, 745]
b = [a] * 10
c = test(b)
assert b == c
assert b is not c
assert c[0] is c[1] and c[1] is c[2]

s1 = 'hello'
s2 = 'world'
a = [s1, s2] * 10
b = test(a)
assert b == a
assert b is not a
assert b[0] is b[2]
assert b[1] is b[3]

from pkpy import TValue

class Base(TValue[int]):
    def __eq__(self, other):
        return self.value == other.value
    
    def __ne__(self, other):
        return self.value != other.value
    
class TVal(Base): pass # type: ignore

test(TVal(1))

old_bytes = pkl.dumps(TVal(1))
print(old_bytes)

# re-define the class so it will have a new type id
class TVal(Base): pass
# see if we can still load the old data
decoded = pkl.loads(old_bytes)
assert decoded == TVal(1)
print(pkl.dumps(decoded))

# test array2d with TValue
a = array2d[TVal].fromlist([
    [TVal(1), TVal(2)],
    [TVal(3), 1]])
test(a)

# test __reduce__

class A:
    def __init__(self, seed):
        self.seed = seed
        self.x = seed
        self.y = seed + 1
        self.z = seed + 2
    def __eq__(self, other):
        return (self.x, self.y, self.z) == (other.x, other.y, other.z)
    def __ne__(self, other):
        return (self.x, self.y, self.z) != (other.x, other.y, other.z)
    def __repr__(self):
        return f"A({self.seed}, x={self.x}, y={self.y}, z={self.z})"
    def __reduce__(self):
        print('__reduce__() called')
        return A, (self.seed,)

test([A(1)]*10)

class Simple:
    def __init__(self, x):
        self.field1 = x
        self.field2 = [...]
    def __eq__(self, other): return self.field1 == other.field1
    def __ne__(self, other): return self.field1 != other.field1

test(Simple(1))
test([Simple(2)]*10)

from dataclasses import dataclass

@dataclass
class Data:
    a: int
    b: str = '2'
    c: float = 3.0

test(Data(1))

exit()

from pickle import dumps, loads, _wrap, _unwrap

def test(x):
    y = dumps(x)
    # print(y.decode())
    ok = x == loads(y)
    if not ok:
        _0 = _wrap(x)
        _1 = _unwrap(_0)
        print('='*50)
        print(_0)
        print('-'*50)
        print(_1)
        print('='*50)
        assert False

test(1)
test(1.0)
test("hello")
test(True)
test(False)
test(None)

test([1, 2, 3])
test((1, 2, 3))
test({1: 2, 3: 4})

class Foo:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Foo):
            return False
        return self.x == __value.x and self.y == __value.y
    
    def __repr__(self) -> str:
        return f"Foo({self.x}, {self.y})"
    
test(Foo(1, 2))
test(Foo([1, True], 'c'))

from vmath import vec2

test(vec2(1, 2))

a = {1, 2, 3, 4}
test(a)

a = bytes([1, 2, 3, 4])
test(a)

a = [1, 2]
d = {'k': a, 'j': a}
c = loads(dumps(d))

assert c['k'] is c['j']
assert c == d

# test circular references
from collections import deque

a = deque([1, 2, 3])
test(a)

a = [int, float, Foo]
test(a)

# test function
def f(x, y):
    return x + y
test(f)

# test @staticmethod
class B:
    @staticmethod
    def f(x, y):
        return x * y
    @classmethod
    def g(cls):
        return cls

class C(B):
    pass

test(B.f)
test(C.f)
test(B.g)
test(C.g)



import traceback
import sys

if sys.argv[0].endswith('.pyc'):
    exit()

try:
    a = {'123': 4}
    b = a[6]
except KeyError:
    actual = traceback.format_exc()

expected = '''Traceback (most recent call last):
  File "tests/802_traceback.py", line 9
    b = a[6]
KeyError: 6'''

if actual != expected:
    print('--- ACTUAL RESULT -----')
    print(actual)
    print('--- EXPECTED RESULT ---')
    print(expected)
    exit(1)


import sys

class vec2:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __add__(self, other):
        return vec2(self.x + other.x, self.y + other.y)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y
    
    def __ne__(self, other):
        return not self == other

x = vec2(0, 0)
for i in range(10000000):
    x += vec2(1, 1)

assert x == vec2(10000000, 10000000)



from inspect import isgeneratorfunction

def f(a, b):
    return a + b

assert not isgeneratorfunction(f)

def g(a, b):
    yield a
    yield b

assert isgeneratorfunction(g)

class A:
    @staticmethod
    def non_gen(a, b):
        return a + b
    @staticmethod
    def gen(a, b):
        yield a
        yield b
    @classmethod
    def non_gen_class(cls, a, b):
        return a + b
    @classmethod
    def gen_class(cls, a, b):
        yield a
        yield b
    def not_gen_instance(self, a, b):
        return a + b
    def gen_instance(self, a, b):
        yield a
        yield b

a = A()
assert not isgeneratorfunction(a.non_gen)
assert isgeneratorfunction(a.gen)
assert not isgeneratorfunction(A.non_gen)
assert isgeneratorfunction(A.gen)

assert not isgeneratorfunction(a.non_gen_class)
assert isgeneratorfunction(a.gen_class)
assert not isgeneratorfunction(A.non_gen_class)
assert isgeneratorfunction(A.gen_class)

assert not isgeneratorfunction(a.not_gen_instance)
assert isgeneratorfunction(a.gen_instance)
assert not isgeneratorfunction(A.not_gen_instance)
assert isgeneratorfunction(A.gen_instance)



a = {
    'a': 1,
    'b': 2,
    'c': None,
    'd': [1, 2, 3],
    'e': {
        'a': 100,
        'b': 2.5,
        'c': None,
        'd': [142, 2785, 39767],
    },
    "f": 'This is a string',
    'g': [True, False, None],
    'h': False
}

import json

assert json.loads("1") == 1
assert json.loads('"1"') == "1"
assert json.loads("0.0") == 0.0
assert json.loads("[1, 2]") == [1, 2]
assert json.loads("null") == None
assert json.loads("true") == True
assert json.loads("false") == False
assert json.loads("{}") == {}

# assert json.loads(b"false") == False

_j = json.dumps(a)
_a = json.loads(_j)

for k, v in a.items():
    assert (a[k] == _a[k]), f'{a[k]} != {_a[k]}'

for k, v in _a.items():
    assert (a[k] == _a[k]), f'{a[k]} != {_a[k]}'

b = [1, 2, True, None, False]

_j = json.dumps(b)
_b = json.loads(_j)

assert b == _b

c = 1.0
_j = json.dumps(c)
_c = json.loads(_j)
assert c == _c

d = True
_j = json.dumps(d)
_d = json.loads(_j)
assert d == _d


assert repr((1,)) == '(1,)'
assert repr((1, 2, 3)) == '(1, 2, 3)'
assert repr(tuple()) == '()'
assert json.dumps((1,)) == '[1]'
assert json.dumps((1, 2, 3)) == '[1, 2, 3]'
assert json.dumps(tuple()) == '[]'

assert repr([]) == '[]'
assert repr([1, 2, 3]) == '[1, 2, 3]'
assert repr([1]) == '[1]'
assert json.dumps([]) == '[]'
assert json.dumps([1, 2, 3]) == '[1, 2, 3]'
assert json.dumps([1]) == '[1]'

try:
    json.dumps({1: 2})
    assert False
except TypeError:
    assert True

try:
    json.dumps(type)
    assert False
except TypeError:
    assert True


class A:
    def __init__(self, a, b):
        self.a = a
        self.b = b

a = A(1, ['2', False, None])
assert json.dumps(a.__dict__) in [
    '{"a": 1, "b": ["2", false, null]}',
    '{"b": ["2", false, null], "a": 1}',
]


from typing import overload, Iterator

class _vecF[T]:
    ONE: T
    ZERO: T

    def __add__(self, other: T) -> T: ...
    def __sub__(self, other: T) -> T: ...
    @overload
    def __mul__(self, other: float) -> T: ...
    @overload
    def __mul__(self, other: T) -> T: ...
    def __truediv__(self, other: float) -> T: ...

    def dot(self, other: T) -> float: ...
    def length(self) -> float: ...
    def length_squared(self) -> float: ...
    def normalize(self) -> T: ...

    # dummy iter for unpacking
    def __iter__(self) -> Iterator[float]: ...

class _vecI[T]:
    ONE: T
    ZERO: T

    def __add__(self, other: T) -> T: ...
    def __sub__(self, other: T) -> T: ...
    @overload
    def __mul__(self, other: int) -> T: ...
    @overload
    def __mul__(self, other: T) -> T: ...
    def __floordiv__(self, other: int) -> T: ...

    def __hash__(self) -> int: ...

    def dot(self, other: T) -> int: ...

    # dummy iter for unpacking
    def __iter__(self) -> Iterator[int]: ...


class vec2(_vecF['vec2']):
    LEFT: vec2
    RIGHT: vec2
    UP: vec2
    DOWN: vec2

    @property
    def x(self) -> float: ...
    @property
    def y(self) -> float: ...

    def with_x(self, x: float) -> vec2: ...
    def with_y(self, y: float) -> vec2: ...
    def with_z(self, z: float) -> vec3: ...

    @overload
    def __init__(self, x: float, y: float) -> None: ...
    @overload
    def __init__(self, xy: vec2i) -> None: ...

    def rotate(self, radians: float) -> vec2:
        """Rotate the vector by `radians`.

        + If y axis is top to bottom, positive value means clockwise (default)
        + If y axis is bottom to top, positive value means counter-clockwise
        """

    @staticmethod
    def angle(__from: vec2, __to: vec2) -> float:
        """Return the angle in radians between vectors `from` and `to`.

        The result range is `[-pi, pi]`.
        
        + If y axis is top to bottom, positive value means clockwise (default)
        + If y axis is bottom to top, positive value means counter-clockwise
        """

    @staticmethod
    def smooth_damp(current: vec2, target: vec2, current_velocity: vec2, smooth_time: float, max_speed: float, delta_time: float) -> tuple[vec2, vec2]:
        """Smoothly change a vector towards a desired goal over time.

        Returns a new value that is closer to the target and current velocity.
        """


class mat3x3:
    def __init__(self, _11, _12, _13, _21, _22, _23, _31, _32, _33) -> None: ...

    def __getitem__(self, index: tuple[int, int]) -> float: ...
    def __setitem__(self, index: tuple[int, int], value: float) -> None: ...

    @overload
    def __matmul__(self, other: mat3x3) -> mat3x3: ...
    @overload
    def __matmul__(self, other: vec3) -> vec3: ...

    def __invert__(self) -> mat3x3: ...

    def matmul(self, other: mat3x3, out: mat3x3) -> mat3x3 | None: ...
    def determinant(self) -> float: ...

    def copy(self) -> mat3x3: ...
    def inverse(self) -> mat3x3: ...

    def copy_(self, other: mat3x3) -> None: ...
    def inverse_(self) -> None: ...

    @staticmethod
    def zeros() -> mat3x3: ...
    @staticmethod
    def identity() -> mat3x3: ...

    # affine transformations
    @staticmethod
    def trs(t: vec2, r: float, s: vec2) -> mat3x3: ...

    def copy_trs_(self, t: vec2, r: float, s: vec2) -> None: ...

    def t(self) -> vec2: ...
    def r(self) -> float: ...
    def s(self) -> vec2: ...

    def transform_point(self, p: vec2) -> vec2: ...
    def transform_vector(self, v: vec2) -> vec2: ...


class vec2i(_vecI['vec2i']):
    LEFT: vec2i
    RIGHT: vec2i
    UP: vec2i
    DOWN: vec2i

    @property
    def x(self) -> int: ...
    @property
    def y(self) -> int: ...

    def with_x(self, x: int) -> vec2i: ...
    def with_y(self, y: int) -> vec2i: ...

    def __init__(self, x: int, y: int) -> None: ...


class vec3i(_vecI['vec3i']):
    @property
    def x(self) -> int: ...
    @property
    def y(self) -> int: ...
    @property
    def z(self) -> int: ...

    def with_x(self, x: int) -> vec3i: ...
    def with_y(self, y: int) -> vec3i: ...
    def with_z(self, z: int) -> vec3i: ...

    def __init__(self, x: int, y: int, z: int) -> None: ...

class vec4i(_vecI['vec4i']):
    @property
    def x(self) -> int: ...
    @property
    def y(self) -> int: ...
    @property
    def z(self) -> int: ...
    @property
    def w(self) -> int: ...

    def with_x(self, x: int) -> vec4i: ...
    def with_y(self, y: int) -> vec4i: ...
    def with_z(self, z: int) -> vec4i: ...
    def with_w(self, w: int) -> vec4i: ...

    def __init__(self, x: int, y: int, z: int, w: int) -> None: ...


class vec3(_vecF['vec3']):
    @property
    def x(self) -> float: ...
    @property
    def y(self) -> float: ...
    @property
    def z(self) -> float: ...

    @property
    def xy(self) -> vec2: ...

    def with_x(self, x: float) -> vec3: ...
    def with_y(self, y: float) -> vec3: ...
    def with_z(self, z: float) -> vec3: ...
    def with_xy(self, xy: vec2) -> vec3: ...

    @overload
    def __init__(self, x: float, y: float, z: float) -> None: ...
    @overload
    def __init__(self, xyz: vec3i) -> None: ...


# Color32
class color32:
    def __new__(cls, r: int, g: int, b: int, a: int) -> 'color32': ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...

    @property
    def r(self) -> int: ...
    @property
    def g(self) -> int: ...
    @property
    def b(self) -> int: ...
    @property
    def a(self) -> int: ...

    def with_r(self, r: int) -> 'color32': ...
    def with_g(self, g: int) -> 'color32': ...
    def with_b(self, b: int) -> 'color32': ...
    def with_a(self, a: int) -> 'color32': ...

    @staticmethod
    def from_hex(hex: str) -> 'color32': ...
    @staticmethod
    def from_vec3(vec: vec3) -> 'color32': ...
    @staticmethod
    def from_vec3i(vec: vec3i) -> 'color32': ...

    def to_hex(self) -> str: ...
    def to_vec3(self) -> vec3: ...
    def to_vec3i(self) -> vec3i: ...

    def to_rgb565(self) -> int: ...

    def ansi_fg(self, text: str) -> str: ...
    def ansi_bg(self, text: str) -> str: ...

    @staticmethod
    def alpha_blend(src: color32, dst: color32 | None) -> color32: ...


def rgb(r: int, g: int, b: int) -> color32: ...
def rgba(r: int, g: int, b: int, a: float) -> color32: ...



import random
import builtins

builtins.print = lambda *x: None

def f(x):
  return -x**2+10

def create_init_DNA():
# out: DNA
  ret = random.randint(-100,100)
  return int_to_bin(ret)

def int_to_bin(x):
# in: int_DNA
# out: DNA
  ret = []
  if x >= 0:
    ret.append(0)
  else:
    ret.append(1)
    x = -x
  for i in [4096,2048,1024,512,256,128,64,32,16,8,4,2,1] :
    if x>=i :
      ret.append(1)
      x -= i
    else :
      ret.append(0)
  return ret

def bin_to_int(x):
# in: DNA
# out: int_DNA
  ret = 0
  new_x = x[:]
  flag = -(new_x[0]*2-1)
  mul = 1
  new_x = reversed(new_x)
  new_x.pop()
  for i in new_x :
    ret += flag * i * mul
    mul *= 2
  return ret

def reversed(x):
  ret = []
  for i in range(0,len(x)):
    ret.append(x[-i-1])
  return ret

def create_DNAs(num):
# in: int
# out: DNAs
  ret = []
  for i in range(num):
    ret.append(create_init_DNA())
  return ret

def bins_to_ints(x):
# in: DNAs
# out: int_DNAs
  ret = []
  for i in x:
    ret.append(bin_to_int(i))
  return ret

def create_probabilitys(x):
# in: DNAs
# out: probabilitys
  scores = survival_scores(x)
  mid = []
  ret = []
  sum = 0
  for i in scores:
    sum+=i
  for i in scores:
    mid.append(1/(i/sum))
  sum = 0
  for i in mid:
    sum+=i
  for i in mid:
    ret.append(i/sum)
  return ret

def survival_scores(x):
# in: DNAs
# out: survival_scores
  ret = []
  for i in x:
    ret.append(f(bin_to_int(i)))
  return ret

def choose_DNA(DNAs,probabilitys):
# in: DNAs,probabilitys
# out: DNA
# probabilitys是由若干(0，1)之间的浮点数组成的数组，这些浮点数的和为1
  i = 0 # i记录取出元素的位置
  ran = random.random()
# a=0
  for max in probabilitys :
    if i != 0 :
      min = probabilitys[i-1]
    else :
      min = 0
    if ran < max and ran >= min :
      return DNAs[i]


def next_gen_DNAs(DNAs,num=None):
  num = num or len(DNAs)
  ret = []
  for i in range(num) :
    ret.append(choose_DNA(DNAs,create_probabilitys(DNAs)))


a = create_DNAs(10)
print(a)
print(bins_to_ints(a))
print(survival_scores(a))
print(create_probabilitys(a))


# generate assert test for list

assert [1, 2, 3] == [1, 2, 3]
assert [1, 2, 3] != [1, 2, 4]

# test + *=
assert [1, 2, 3] + [4, 5, 6] == [1, 2, 3, 4, 5, 6]          
assert [1, 2, 3] * 3 == [1, 2, 3, 1, 2, 3, 1, 2, 3]

l = [1,2,3,4]
assert l[2] == 3
assert l[-1] == 4
assert l[:32] == [1,2,3,4]
assert l[32:] == []
assert l[1:4] == [2,3,4]
assert l[-1:-3] == []
assert l[-3:-1] == [2,3]

l1 = [1];l2 = l1;l1.append(2);l3 = [1,1,2]
assert l2[1] == 2
assert l1 == l2
assert l1*3 == [1,2,1,2,1,2]
assert 2*l1 == [1,2,1,2]
assert l3.count(1) == 2

member = ['Tom', 'Sunny', 'Honer', 'Lily']
teacher = [1,2,3]
assert len(member + teacher) == 7
assert member[0] == 'Tom'
assert member[-2] == 'Honer'
assert member[0:3] == ['Tom', 'Sunny', 'Honer']

member.remove('Sunny')
assert member == ['Tom', 'Honer', 'Lily']
member.pop()
assert member == ['Tom', 'Honer']
del member[0]
assert member == ['Honer']
member.append('Jack')
assert member == ['Honer','Jack']
member.extend(teacher)
assert member == ['Honer','Jack',1,2,3]
member.insert(1,'Tom')
assert member == ['Honer','Tom','Jack',1,2,3]
member.clear()
assert member == []
member = teacher.copy()
assert member == [1,2,3]

l = []
l.insert(0, 'l')
l.insert(1, 'l')
l.insert(0, 'h')
l.insert(3, 'o')
l.insert(1, 'e')
assert l == ['h', 'e', 'l', 'l', 'o']
assert l[-2] == 'l'

b = [1, 2, 3]
assert b[
    1
] == 2

assert b[0
] == 1

assert b[0] == 1
assert b[
    0] == 1

a = [1, 2, 3]
assert a.index(2) == 1
assert a.index(1) == 0
assert a.index(3) == 2

assert a.index(2, 1) == 1
assert a.index(1, 0) == 0
assert a.index(2, -2) == 1

a, b = [1, 2]
assert a == 1 and b == 2
assert [1, 2].__getitem__(0) == 1

assert list(range(1, 5)) == [1, 2, 3, 4]
assert list(range(1, 5, 2)) == [1, 3]
assert list(range(5, 1, -1)) == [5, 4, 3, 2]
assert list(range(5, 1, -2)) == [5, 3]

# test sort
a = [8, 2, 4, 2, 9]
assert a.sort() == None
assert (a == [2, 2, 4, 8, 9]), a

a = []
assert a.sort() == None
assert a == []

a = [0, 0, 0, 0, 1, 1, 3, -1]
assert a.sort() == None
assert a == [-1, 0, 0, 0, 0, 1, 1, 3]

a = [3, 2, 2, 1]
a.sort()
assert a == [1, 2, 2, 3]

# test reverse
a = [3, 2, 2, 1]
a.sort(reverse=True)
assert a == [3, 2, 2, 1]

a = [1, 3, 2, 2]
a.sort(reverse=True)
assert a == [3, 2, 2, 1]

# test key
key = lambda x: -x
assert key(1) == -1
a = [1, 3, 2, 2]
a.sort(key=key)
assert a == [3, 2, 2, 1]

a = [1, 3, 2, 2]
a.sort(key=key, reverse=True)
assert a == [1, 2, 2, 3]

# test sorted
a = [8, 2, 4, 2, 9]
assert sorted(a) == [2, 2, 4, 8, 9]
assert sorted(a, reverse=True) == [9, 8, 4, 2, 2]

def key(x): return -x;
assert sorted(a, key=key) == [9, 8, 4, 2, 2]

assert sorted(a, key=key, reverse=True) == [2, 2, 4, 8, 9]
assert a == [8, 2, 4, 2, 9]

# test unpack ex
a, *b = [1,2,3,4]
assert a == 1
assert b == [2,3,4]

a, b, *c = [1, 2]
assert a == 1
assert b == 2
assert c == []

a, b, c, *d = [1, 2, 3, 4, 5]
assert d == [4, 5]

# test repr
assert repr([1, 2, 3]) == "[1, 2, 3]"
assert repr([1, [2, 3], 4]) == "[1, [2, 3], 4]"
assert repr([1, [2, [3, 4]], 5]) == "[1, [2, [3, 4]], 5]"
assert repr([]) == "[]"

# test in and not in
assert 1 in [1, 2, 3]
assert 4 not in [1, 2, 3]

b = [(1, 2), (3, 3), (5, 1)]
b.sort(key=lambda x:x[1])
assert b == [(5, 1), (1, 2), (3,3)]

# test del slice
a = [1, 2, 3, 4]
b = a.copy(); del b[:2]; assert b == [3, 4]
b = a.copy(); del b[1:3]; assert b == [1, 4]
b = a.copy(); del b[2:]; assert b == [1, 2]
b = a.copy(); del b[:-1]; assert b == [4]
b = a.copy(); del b[-1:]; assert b == [1, 2, 3]
b = a.copy(); del b[:]; assert b == []
assert a == [1, 2, 3, 4]

# test extend with iterable
c = [1]
c.extend('123')
assert c == [1, '1', '2', '3']
c.extend(range(1, 6))
assert c == [1, '1', '2', '3', 1, 2, 3, 4, 5]

# test cyclic reference
# a = []
# a.append(0)
# a.append([1, 2, a])

# assert repr(a) == "[0, [1, 2, [...]]]"

# unpacking builder (not supported)
# a = [1, 2, 3]
# b = [*a, 4, 5]
# assert b == [1, 2, 3, 4, 5]

# a = []
# b = [*a, 1, 2, 3, *a, *a]
# assert b == [1, 2, 3]


import datetime


def test_timedelta():
    assert datetime.timedelta(days=1) == datetime.timedelta(days=1)
    assert datetime.timedelta(days=1) != datetime.timedelta(days=2)

def test_date():
    assert datetime.date(2023, 8, 5) == datetime.date(2023, 8, 5)
    assert datetime.date(2023, 8, 5) != datetime.date(2023, 8, 6)
    assert datetime.date(2024, 8, 5) >= datetime.date(2023, 8, 6)
    assert datetime.date(2023, 8, 5) <= datetime.date(2023, 8, 6)
    assert datetime.date(2024, 8, 5) > datetime.date(2023, 8, 6)
    assert datetime.date(2023, 8, 5) < datetime.date(2024, 8, 6)

def test_datetime():
    assert datetime.datetime(
        2023, 8, 5, 12, 0, 0) == datetime.datetime(2023, 8, 5, 12, 0, 0)
    assert datetime.datetime(
        2023, 8, 5, 12, 0, 0) != datetime.datetime(2023, 8, 5, 12, 1, 0)
    assert datetime.datetime(
        2023, 8, 5, 12, 0, 0) >= datetime.datetime(2023, 8, 5, 12, 0, 0)
    assert datetime.datetime(
        2023, 8, 5, 12, 30, 0) > datetime.datetime(2023, 8, 5, 12, 1, 0)
    assert datetime.datetime(
        2023, 8, 5, 12, 0, 0) < datetime.datetime(2023, 8, 5, 12, 1, 0)
    assert datetime.datetime(
        2023, 8, 5, 12, 0, 0) <= datetime.datetime(2023, 8, 5, 12, 1, 0)

test_timedelta()
test_date()
test_datetime()



# 未完全测试准确性-----------------------------------------------
#       116: 1529:    bind_property(_t(tp_slice), "start", [](VM* vm, ArgsView args){
#     #####: 1530:        return CAST(Slice&, args[0]).start;
#         -: 1531:    });
#       116: 1532:    bind_property(_t(tp_slice), "stop", [](VM* vm, ArgsView args){
#     #####: 1533:        return CAST(Slice&, args[0]).stop;
#         -: 1534:    });
#       116: 1535:    bind_property(_t(tp_slice), "step", [](VM* vm, ArgsView args){
#     #####: 1536:        return CAST(Slice&, args[0]).step;
#         -: 1537:    });
s = slice(1, 2, 3)
assert type(s) is slice
assert s.start == 1
assert s.stop == 2
assert s.step == 3

# 未完全测试准确性-----------------------------------------------
# test slice.__repr__
assert type(repr(slice(1,1,1))) is str


class A():
    def __init__(self):
        self.a = 10
    def method(self):
        pass


my_namedict = A().__dict__

try:
    hash(my_namedict)
    print('未能拦截错误, 在测试 namedict.__hash__')
    exit(1)
except TypeError:
    pass

a = hash(object())  # object is hashable
a = hash(A())       # A is hashable
class B:
    def __eq__(self, o): return True
    def __ne__(self, o): return False

try:
    hash(B())
    print('未能拦截错误, 在测试 B.__hash__')
    exit(1)
except TypeError:
    pass

# 未完全测试准确性-----------------------------------------------
# test namedict.__repr__:
class A():
    def __init__(self):
        self.a = 10
    def method(self):
        pass


my_namedict = A().__dict__
assert type(repr(my_namedict)) is str


# /************ dict ************/
# 未完全测试准确性-----------------------------------------------
# test dict:
assert type(dict([(1,2)])) is dict

try:
    dict([(1, 2, 3)])
    print('未能拦截错误, 在测试 dict')
    exit(1)
except ValueError:
    pass

try:
    dict([(1, 2)], 1)
    print('未能拦截错误, 在测试 dict')
    exit(1)
except TypeError:
    pass

try:
    hash(dict([(1,2)]))
    print('未能拦截错误, 在测试 dict.__hash__')
    exit(1)
except TypeError:
    pass

# test dict.__iter__
for k in {1:2, 2:3, 3:4}.keys():
    assert k in [1,2,3]

# 未完全测试准确性-----------------------------------------------
# test dict.get

assert {1:2, 3:4}.get(1) == 2
assert {1:2, 3:4}.get(2) is None
assert {1:2, 3:4}.get(20, 100) == 100

try:
    {1:2, 3:4}.get(1,1, 1)
    print('未能拦截错误, 在测试 dict.get')
    exit(1)
except TypeError:
    pass

# 未完全测试准确性-----------------------------------------------
# test dict.__repr__
assert type(repr({1:2, 3:4})) is str

# /************ property ************/
class A():
    def __init__(self):
        self._name = '123'

    @property
    def value(self):
        return 2

    def get_name(self):
        '''
        doc string 1
        '''
        return self._name

    def set_name(self, val):
        '''
        doc string 2
        '''
        self._name = val

assert A().value == 2

A.name = property(A.get_name, A.set_name)

class Vector2:
    def __init__(self) -> None:
        self._x = 0

    @property
    def x(self):
        return self._x
    
    @x.setter
    def x(self, val):
        self._x = val

v = Vector2()
assert v.x == 0
v.x = 10
assert v.x == 10

# function.__doc__
def aaa():
    '12345'
    pass
assert aaa.__doc__ == '12345'

# test callable
assert callable(lambda: 1) is True          # function
assert callable(1) is False                 # int
assert callable(object) is True             # type
assert callable(object()) is False
assert callable([].append) is True      # bound method
assert callable([].__getitem__) is True # bound method

class A:
    def __init__(self):
        pass

    def __call__(self):
        pass

    @staticmethod
    def staticmethod():
        pass

    @classmethod
    def classmethod(cls):
        pass

assert callable(A) is True      # type
assert callable(A()) is True    # instance with __call__
assert callable(A.__call__) is True  # bound method
assert callable(A.__init__) is True  # bound method
assert callable(print) is True  # builtin function
assert callable(isinstance) is True  # builtin function
assert callable(A.staticmethod) is True  # staticmethod
assert callable(A.classmethod) is True  # classmethod

assert id(0) is None
assert id(2**62) is None

# test issubclass
assert issubclass(int, int) is True
assert issubclass(int, object) is True
assert issubclass(object, int) is False
assert issubclass(object, object) is True
assert issubclass(int, type) is False
assert issubclass(type, type) is True
assert issubclass(float, int) is False


def f(a, b):
    c = a
    del a
    return sum([b, c])

assert f(1, 2) == 3

# /************ module time ************/
import time
# test time.time
assert type(time.time()) is float

local_t = time.localtime()
assert type(local_t.tm_year) is int
assert type(local_t.tm_mon) is int
assert type(local_t.tm_mday) is int
assert type(local_t.tm_hour) is int
assert type(local_t.tm_min) is int
assert type(local_t.tm_sec) is int
assert type(local_t.tm_wday) is int
assert type(local_t.tm_yday) is int
assert type(local_t.tm_isdst) is int

# test time.sleep
time.sleep(0.1)
# test time.localtime
assert type(time.localtime()) is time.struct_time

# test min/max
assert min(1, 2) == 1
assert min(1, 2, 3) == 1
assert min([1, 2]) == 1
assert min([1, 2], key=lambda x: -x) == 2

assert max(1, 2) == 2
assert max(1, 2, 3) == 3
assert max([1, 2]) == 2
assert max([1, 2, 3], key=lambda x: -x) == 1

assert min([
    (3, 1),
    (1, 2),
    (1, 3),
    (1, 4),
]) == (1, 2)

assert min(1, 2) == 1
assert max(1, 2) == 2

def fn(): pass
assert repr(fn).startswith('<function fn at')

assert repr(round) == '<nativefunc object>'


a = '12345'
assert a.encode() == b'12345'

# test add
assert b'123' + b'456' == b'123456'
assert b'' + b'123' == b'123'
assert b'123' + b'' == b'123'
assert b'' + b'' == b''

assert b'\xff\xee' != b'1234'
assert b'\xff\xee' == b'\xff\xee'
assert len(b'\xff\xee') == 2
assert len(b'') == 0

a = '测试123'
assert a == a.encode().decode()

assert chr(0) == '\x00'
assert ord('\x00') == 0

# test slice
s = b"football"
q = b"abcd"
r = b"zoo"
t = b"this is string example....wow!!!"
assert s[0] == ord('f')
assert s[1:4] == b'oot'
assert s[:-1] == b'footbal'
assert s[:10] == b'football'
assert s[-3] == ord('a')
assert t[-5:] == b'ow!!!'
assert t[3:-3] == b's is string example....wow'

a = b"Hello, World!"
assert a[::-1] == b"!dlroW ,olleH"
assert a[::2] == b"Hlo ol!"
assert a[2:5:2] == b"lo"
assert a[5:2:-1] == b",ol"
assert a[5:2:-2] == b",l"

assert bytes() == b''
assert bytes((65,)) == b'A'
assert bytes([0, 1, 2, 3]) == b'\x00\x01\x02\x03'


from typing import overload
from stdc import intptr

class gpio_config:
    direction: int #  (gpio_direction_t)
    edge: int #  (gpio_edge_t)
    event_clock: int #  (gpio_event_clock_t)
    debounce_us: int #  (uint32_t)
    bias: int #  (gpio_bias_t)
    drive: int #  (gpio_drive_t)
    inverted: bool #  (bool)
    label: str #  (const char*)
    
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, direction: int, edge: int, event_clock: int, debounce_us: int, bias: int, drive: int, inverted: bool, label: str): ...
    
class spi_msg:
    txbuf: intptr #  (const uint8_t*)
    rxbuf: intptr #  (uint8_t*)
    len: int #  (size_t)
    deselect: bool #  (bool)
    deselect_delay_us: int #  (uint16_t)
    word_delay_us: int #  (uint8_t)
    
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, txbuf: intptr, rxbuf: intptr, len: int, deselect: bool, deselect_delay_us: int, word_delay_us: int): ...
    
class periphery_version:
    major: int #  (unsigned)
    minor: int #  (unsigned)
    patch: int #  (unsigned)
    commit_id: str #  (const char*)
    
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, major: int, minor: int, patch: int, commit_id: str): ...
    
def gpio_new() -> intptr:
    """Wraps `gpio_t* gpio_new()`"""

def gpio_open(gpio: intptr, path: str, line: int, direction: int, /) -> int:
    """Wraps `int gpio_open(gpio_t* gpio, const char* path, unsigned line, gpio_direction_t direction)`"""

def gpio_open_name(gpio: intptr, path: str, name: str, direction: int, /) -> int:
    """Wraps `int gpio_open_name(gpio_t* gpio, const char* path, const char* name, gpio_direction_t direction)`"""

def gpio_open_advanced(gpio: intptr, path: str, line: int, config: intptr, /) -> int:
    """Wraps `int gpio_open_advanced(gpio_t* gpio, const char* path, unsigned line, const gpio_config_t* config)`"""

def gpio_open_name_advanced(gpio: intptr, path: str, name: str, config: intptr, /) -> int:
    """Wraps `int gpio_open_name_advanced(gpio_t* gpio, const char* path, const char* name, const gpio_config_t* config)`"""

def gpio_open_sysfs(gpio: intptr, line: int, direction: int, /) -> int:
    """Wraps `int gpio_open_sysfs(gpio_t* gpio, unsigned line, gpio_direction_t direction)`"""

def gpio_read(gpio: intptr, value: intptr, /) -> int:
    """Wraps `int gpio_read(gpio_t* gpio, bool* value)`"""

def gpio_write(gpio: intptr, value: bool, /) -> int:
    """Wraps `int gpio_write(gpio_t* gpio, bool value)`"""

def gpio_poll(gpio: intptr, timeout_ms: int, /) -> int:
    """Wraps `int gpio_poll(gpio_t* gpio, int timeout_ms)`"""

def gpio_close(gpio: intptr, /) -> int:
    """Wraps `int gpio_close(gpio_t* gpio)`"""

def gpio_free(gpio: intptr, /) -> None:
    """Wraps `void gpio_free(gpio_t* gpio)`"""

def gpio_read_event(gpio: intptr, edge: intptr, timestamp: intptr, /) -> int:
    """Wraps `int gpio_read_event(gpio_t* gpio, gpio_edge_t* edge, uint64_t* timestamp)`"""

def gpio_poll_multiple(gpios: intptr, count: int, timeout_ms: int, gpios_ready: intptr, /) -> int:
    """Wraps `int gpio_poll_multiple(gpio_t** gpios, size_t count, int timeout_ms, bool* gpios_ready)`"""

def gpio_get_direction(gpio: intptr, direction: intptr, /) -> int:
    """Wraps `int gpio_get_direction(gpio_t* gpio, gpio_direction_t* direction)`"""

def gpio_get_edge(gpio: intptr, edge: intptr, /) -> int:
    """Wraps `int gpio_get_edge(gpio_t* gpio, gpio_edge_t* edge)`"""

def gpio_get_event_clock(gpio: intptr, event_clock: intptr, /) -> int:
    """Wraps `int gpio_get_event_clock(gpio_t* gpio, gpio_event_clock_t* event_clock)`"""

def gpio_get_debounce_us(gpio: intptr, debounce_us: intptr, /) -> int:
    """Wraps `int gpio_get_debounce_us(gpio_t* gpio, uint32_t* debounce_us)`"""

def gpio_get_bias(gpio: intptr, bias: intptr, /) -> int:
    """Wraps `int gpio_get_bias(gpio_t* gpio, gpio_bias_t* bias)`"""

def gpio_get_drive(gpio: intptr, drive: intptr, /) -> int:
    """Wraps `int gpio_get_drive(gpio_t* gpio, gpio_drive_t* drive)`"""

def gpio_get_inverted(gpio: intptr, inverted: intptr, /) -> int:
    """Wraps `int gpio_get_inverted(gpio_t* gpio, bool* inverted)`"""

def gpio_set_direction(gpio: intptr, direction: int, /) -> int:
    """Wraps `int gpio_set_direction(gpio_t* gpio, gpio_direction_t direction)`"""

def gpio_set_edge(gpio: intptr, edge: int, /) -> int:
    """Wraps `int gpio_set_edge(gpio_t* gpio, gpio_edge_t edge)`"""

def gpio_set_event_clock(gpio: intptr, event_clock: int, /) -> int:
    """Wraps `int gpio_set_event_clock(gpio_t* gpio, gpio_event_clock_t event_clock)`"""

def gpio_set_debounce_us(gpio: intptr, debounce_us: int, /) -> int:
    """Wraps `int gpio_set_debounce_us(gpio_t* gpio, uint32_t debounce_us)`"""

def gpio_set_bias(gpio: intptr, bias: int, /) -> int:
    """Wraps `int gpio_set_bias(gpio_t* gpio, gpio_bias_t bias)`"""

def gpio_set_drive(gpio: intptr, drive: int, /) -> int:
    """Wraps `int gpio_set_drive(gpio_t* gpio, gpio_drive_t drive)`"""

def gpio_set_inverted(gpio: intptr, inverted: bool, /) -> int:
    """Wraps `int gpio_set_inverted(gpio_t* gpio, bool inverted)`"""

def gpio_line(gpio: intptr, /) -> int:
    """Wraps `unsigned gpio_line(gpio_t* gpio)`"""

def gpio_fd(gpio: intptr, /) -> int:
    """Wraps `int gpio_fd(gpio_t* gpio)`"""

def gpio_name(gpio: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int gpio_name(gpio_t* gpio, char* str, size_t len)`"""

def gpio_label(gpio: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int gpio_label(gpio_t* gpio, char* str, size_t len)`"""

def gpio_chip_fd(gpio: intptr, /) -> int:
    """Wraps `int gpio_chip_fd(gpio_t* gpio)`"""

def gpio_chip_name(gpio: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int gpio_chip_name(gpio_t* gpio, char* str, size_t len)`"""

def gpio_chip_label(gpio: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int gpio_chip_label(gpio_t* gpio, char* str, size_t len)`"""

def gpio_tostring(gpio: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int gpio_tostring(gpio_t* gpio, char* str, size_t len)`"""

def gpio_errno(gpio: intptr, /) -> int:
    """Wraps `int gpio_errno(gpio_t* gpio)`"""

def gpio_errmsg(gpio: intptr, /) -> str:
    """Wraps `const char* gpio_errmsg(gpio_t* gpio)`"""

def led_new() -> intptr:
    """Wraps `led_t* led_new()`"""

def led_open(led: intptr, name: str, /) -> int:
    """Wraps `int led_open(led_t* led, const char* name)`"""

def led_read(led: intptr, value: intptr, /) -> int:
    """Wraps `int led_read(led_t* led, bool* value)`"""

def led_write(led: intptr, value: bool, /) -> int:
    """Wraps `int led_write(led_t* led, bool value)`"""

def led_close(led: intptr, /) -> int:
    """Wraps `int led_close(led_t* led)`"""

def led_free(led: intptr, /) -> None:
    """Wraps `void led_free(led_t* led)`"""

def led_get_brightness(led: intptr, brightness: intptr, /) -> int:
    """Wraps `int led_get_brightness(led_t* led, unsigned* brightness)`"""

def led_get_max_brightness(led: intptr, max_brightness: intptr, /) -> int:
    """Wraps `int led_get_max_brightness(led_t* led, unsigned* max_brightness)`"""

def led_get_trigger(led: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int led_get_trigger(led_t* led, char* str, size_t len)`"""

def led_get_triggers_entry(led: intptr, index: int, str: intptr, len: int, /) -> int:
    """Wraps `int led_get_triggers_entry(led_t* led, unsigned index, char* str, size_t len)`"""

def led_get_triggers_count(led: intptr, count: intptr, /) -> int:
    """Wraps `int led_get_triggers_count(led_t* led, unsigned* count)`"""

def led_set_brightness(led: intptr, brightness: int, /) -> int:
    """Wraps `int led_set_brightness(led_t* led, unsigned brightness)`"""

def led_set_trigger(led: intptr, trigger: str, /) -> int:
    """Wraps `int led_set_trigger(led_t* led, const char* trigger)`"""

def led_name(led: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int led_name(led_t* led, char* str, size_t len)`"""

def led_tostring(led: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int led_tostring(led_t* led, char* str, size_t len)`"""

def led_errno(led: intptr, /) -> int:
    """Wraps `int led_errno(led_t* led)`"""

def led_errmsg(led: intptr, /) -> str:
    """Wraps `const char* led_errmsg(led_t* led)`"""

def mmio_new() -> intptr:
    """Wraps `mmio_t* mmio_new()`"""

def mmio_open(mmio: intptr, base: int, size: int, /) -> int:
    """Wraps `int mmio_open(mmio_t* mmio, uintptr_t base, size_t size)`"""

def mmio_open_advanced(mmio: intptr, base: int, size: int, path: str, /) -> int:
    """Wraps `int mmio_open_advanced(mmio_t* mmio, uintptr_t base, size_t size, const char* path)`"""

def mmio_ptr(mmio: intptr, /) -> intptr:
    """Wraps `void* mmio_ptr(mmio_t* mmio)`"""

def mmio_read64(mmio: intptr, offset: int, value: intptr, /) -> int:
    """Wraps `int mmio_read64(mmio_t* mmio, uintptr_t offset, uint64_t* value)`"""

def mmio_read32(mmio: intptr, offset: int, value: intptr, /) -> int:
    """Wraps `int mmio_read32(mmio_t* mmio, uintptr_t offset, uint32_t* value)`"""

def mmio_read16(mmio: intptr, offset: int, value: intptr, /) -> int:
    """Wraps `int mmio_read16(mmio_t* mmio, uintptr_t offset, uint16_t* value)`"""

def mmio_read8(mmio: intptr, offset: int, value: intptr, /) -> int:
    """Wraps `int mmio_read8(mmio_t* mmio, uintptr_t offset, uint8_t* value)`"""

def mmio_read(mmio: intptr, offset: int, buf: intptr, len: int, /) -> int:
    """Wraps `int mmio_read(mmio_t* mmio, uintptr_t offset, uint8_t* buf, size_t len)`"""

def mmio_write64(mmio: intptr, offset: int, value: int, /) -> int:
    """Wraps `int mmio_write64(mmio_t* mmio, uintptr_t offset, uint64_t value)`"""

def mmio_write32(mmio: intptr, offset: int, value: int, /) -> int:
    """Wraps `int mmio_write32(mmio_t* mmio, uintptr_t offset, uint32_t value)`"""

def mmio_write16(mmio: intptr, offset: int, value: int, /) -> int:
    """Wraps `int mmio_write16(mmio_t* mmio, uintptr_t offset, uint16_t value)`"""

def mmio_write8(mmio: intptr, offset: int, value: int, /) -> int:
    """Wraps `int mmio_write8(mmio_t* mmio, uintptr_t offset, uint8_t value)`"""

def mmio_write(mmio: intptr, offset: int, buf: intptr, len: int, /) -> int:
    """Wraps `int mmio_write(mmio_t* mmio, uintptr_t offset, const uint8_t* buf, size_t len)`"""

def mmio_close(mmio: intptr, /) -> int:
    """Wraps `int mmio_close(mmio_t* mmio)`"""

def mmio_free(mmio: intptr, /) -> None:
    """Wraps `void mmio_free(mmio_t* mmio)`"""

def mmio_base(mmio: intptr, /) -> int:
    """Wraps `uintptr_t mmio_base(mmio_t* mmio)`"""

def mmio_size(mmio: intptr, /) -> int:
    """Wraps `size_t mmio_size(mmio_t* mmio)`"""

def mmio_tostring(mmio: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int mmio_tostring(mmio_t* mmio, char* str, size_t len)`"""

def mmio_errno(mmio: intptr, /) -> int:
    """Wraps `int mmio_errno(mmio_t* mmio)`"""

def mmio_errmsg(mmio: intptr, /) -> str:
    """Wraps `const char* mmio_errmsg(mmio_t* mmio)`"""

def pwm_new() -> intptr:
    """Wraps `pwm_t* pwm_new()`"""

def pwm_open(pwm: intptr, chip: int, channel: int, /) -> int:
    """Wraps `int pwm_open(pwm_t* pwm, unsigned chip, unsigned channel)`"""

def pwm_enable(pwm: intptr, /) -> int:
    """Wraps `int pwm_enable(pwm_t* pwm)`"""

def pwm_disable(pwm: intptr, /) -> int:
    """Wraps `int pwm_disable(pwm_t* pwm)`"""

def pwm_close(pwm: intptr, /) -> int:
    """Wraps `int pwm_close(pwm_t* pwm)`"""

def pwm_free(pwm: intptr, /) -> None:
    """Wraps `void pwm_free(pwm_t* pwm)`"""

def pwm_get_enabled(pwm: intptr, enabled: intptr, /) -> int:
    """Wraps `int pwm_get_enabled(pwm_t* pwm, bool* enabled)`"""

def pwm_get_period_ns(pwm: intptr, period_ns: intptr, /) -> int:
    """Wraps `int pwm_get_period_ns(pwm_t* pwm, uint64_t* period_ns)`"""

def pwm_get_duty_cycle_ns(pwm: intptr, duty_cycle_ns: intptr, /) -> int:
    """Wraps `int pwm_get_duty_cycle_ns(pwm_t* pwm, uint64_t* duty_cycle_ns)`"""

def pwm_get_period(pwm: intptr, period: intptr, /) -> int:
    """Wraps `int pwm_get_period(pwm_t* pwm, double* period)`"""

def pwm_get_duty_cycle(pwm: intptr, duty_cycle: intptr, /) -> int:
    """Wraps `int pwm_get_duty_cycle(pwm_t* pwm, double* duty_cycle)`"""

def pwm_get_frequency(pwm: intptr, frequency: intptr, /) -> int:
    """Wraps `int pwm_get_frequency(pwm_t* pwm, double* frequency)`"""

def pwm_get_polarity(pwm: intptr, polarity: intptr, /) -> int:
    """Wraps `int pwm_get_polarity(pwm_t* pwm, pwm_polarity_t* polarity)`"""

def pwm_set_enabled(pwm: intptr, enabled: bool, /) -> int:
    """Wraps `int pwm_set_enabled(pwm_t* pwm, bool enabled)`"""

def pwm_set_period_ns(pwm: intptr, period_ns: int, /) -> int:
    """Wraps `int pwm_set_period_ns(pwm_t* pwm, uint64_t period_ns)`"""

def pwm_set_duty_cycle_ns(pwm: intptr, duty_cycle_ns: int, /) -> int:
    """Wraps `int pwm_set_duty_cycle_ns(pwm_t* pwm, uint64_t duty_cycle_ns)`"""

def pwm_set_period(pwm: intptr, period: float, /) -> int:
    """Wraps `int pwm_set_period(pwm_t* pwm, double period)`"""

def pwm_set_duty_cycle(pwm: intptr, duty_cycle: float, /) -> int:
    """Wraps `int pwm_set_duty_cycle(pwm_t* pwm, double duty_cycle)`"""

def pwm_set_frequency(pwm: intptr, frequency: float, /) -> int:
    """Wraps `int pwm_set_frequency(pwm_t* pwm, double frequency)`"""

def pwm_set_polarity(pwm: intptr, polarity: int, /) -> int:
    """Wraps `int pwm_set_polarity(pwm_t* pwm, pwm_polarity_t polarity)`"""

def pwm_chip(pwm: intptr, /) -> int:
    """Wraps `unsigned pwm_chip(pwm_t* pwm)`"""

def pwm_channel(pwm: intptr, /) -> int:
    """Wraps `unsigned pwm_channel(pwm_t* pwm)`"""

def pwm_tostring(pwm: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int pwm_tostring(pwm_t* pwm, char* str, size_t len)`"""

def pwm_errno(pwm: intptr, /) -> int:
    """Wraps `int pwm_errno(pwm_t* pwm)`"""

def pwm_errmsg(pwm: intptr, /) -> str:
    """Wraps `const char* pwm_errmsg(pwm_t* pwm)`"""

def serial_new() -> intptr:
    """Wraps `serial_t* serial_new()`"""

def serial_open(serial: intptr, path: str, baudrate: int, /) -> int:
    """Wraps `int serial_open(serial_t* serial, const char* path, uint32_t baudrate)`"""

def serial_open_advanced(serial: intptr, path: str, baudrate: int, databits: int, parity: int, stopbits: int, xonxoff: bool, rtscts: bool, /) -> int:
    """Wraps `int serial_open_advanced(serial_t* serial, const char* path, uint32_t baudrate, unsigned databits, serial_parity_t parity, unsigned stopbits, bool xonxoff, bool rtscts)`"""

def serial_read(serial: intptr, buf: intptr, len: int, timeout_ms: int, /) -> int:
    """Wraps `int serial_read(serial_t* serial, uint8_t* buf, size_t len, int timeout_ms)`"""

def serial_write(serial: intptr, buf: intptr, len: int, /) -> int:
    """Wraps `int serial_write(serial_t* serial, const uint8_t* buf, size_t len)`"""

def serial_flush(serial: intptr, /) -> int:
    """Wraps `int serial_flush(serial_t* serial)`"""

def serial_input_waiting(serial: intptr, count: intptr, /) -> int:
    """Wraps `int serial_input_waiting(serial_t* serial, unsigned* count)`"""

def serial_output_waiting(serial: intptr, count: intptr, /) -> int:
    """Wraps `int serial_output_waiting(serial_t* serial, unsigned* count)`"""

def serial_poll(serial: intptr, timeout_ms: int, /) -> int:
    """Wraps `int serial_poll(serial_t* serial, int timeout_ms)`"""

def serial_close(serial: intptr, /) -> int:
    """Wraps `int serial_close(serial_t* serial)`"""

def serial_free(serial: intptr, /) -> None:
    """Wraps `void serial_free(serial_t* serial)`"""

def serial_get_baudrate(serial: intptr, baudrate: intptr, /) -> int:
    """Wraps `int serial_get_baudrate(serial_t* serial, uint32_t* baudrate)`"""

def serial_get_databits(serial: intptr, databits: intptr, /) -> int:
    """Wraps `int serial_get_databits(serial_t* serial, unsigned* databits)`"""

def serial_get_parity(serial: intptr, parity: intptr, /) -> int:
    """Wraps `int serial_get_parity(serial_t* serial, serial_parity_t* parity)`"""

def serial_get_stopbits(serial: intptr, stopbits: intptr, /) -> int:
    """Wraps `int serial_get_stopbits(serial_t* serial, unsigned* stopbits)`"""

def serial_get_xonxoff(serial: intptr, xonxoff: intptr, /) -> int:
    """Wraps `int serial_get_xonxoff(serial_t* serial, bool* xonxoff)`"""

def serial_get_rtscts(serial: intptr, rtscts: intptr, /) -> int:
    """Wraps `int serial_get_rtscts(serial_t* serial, bool* rtscts)`"""

def serial_get_vmin(serial: intptr, vmin: intptr, /) -> int:
    """Wraps `int serial_get_vmin(serial_t* serial, unsigned* vmin)`"""

def serial_get_vtime(serial: intptr, vtime: intptr, /) -> int:
    """Wraps `int serial_get_vtime(serial_t* serial, float* vtime)`"""

def serial_set_baudrate(serial: intptr, baudrate: int, /) -> int:
    """Wraps `int serial_set_baudrate(serial_t* serial, uint32_t baudrate)`"""

def serial_set_databits(serial: intptr, databits: int, /) -> int:
    """Wraps `int serial_set_databits(serial_t* serial, unsigned databits)`"""

def serial_set_parity(serial: intptr, parity: int, /) -> int:
    """Wraps `int serial_set_parity(serial_t* serial, serial_parity_t parity)`"""

def serial_set_stopbits(serial: intptr, stopbits: int, /) -> int:
    """Wraps `int serial_set_stopbits(serial_t* serial, unsigned stopbits)`"""

def serial_set_xonxoff(serial: intptr, enabled: bool, /) -> int:
    """Wraps `int serial_set_xonxoff(serial_t* serial, bool enabled)`"""

def serial_set_rtscts(serial: intptr, enabled: bool, /) -> int:
    """Wraps `int serial_set_rtscts(serial_t* serial, bool enabled)`"""

def serial_set_vmin(serial: intptr, vmin: int, /) -> int:
    """Wraps `int serial_set_vmin(serial_t* serial, unsigned vmin)`"""

def serial_set_vtime(serial: intptr, vtime: float, /) -> int:
    """Wraps `int serial_set_vtime(serial_t* serial, float vtime)`"""

def serial_fd(serial: intptr, /) -> int:
    """Wraps `int serial_fd(serial_t* serial)`"""

def serial_tostring(serial: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int serial_tostring(serial_t* serial, char* str, size_t len)`"""

def serial_errno(serial: intptr, /) -> int:
    """Wraps `int serial_errno(serial_t* serial)`"""

def serial_errmsg(serial: intptr, /) -> str:
    """Wraps `const char* serial_errmsg(serial_t* serial)`"""

def spi_new() -> intptr:
    """Wraps `spi_t* spi_new()`"""

def spi_open(spi: intptr, path: str, mode: int, max_speed: int, /) -> int:
    """Wraps `int spi_open(spi_t* spi, const char* path, unsigned mode, uint32_t max_speed)`"""

def spi_open_advanced(spi: intptr, path: str, mode: int, max_speed: int, bit_order: int, bits_per_word: int, extra_flags: int, /) -> int:
    """Wraps `int spi_open_advanced(spi_t* spi, const char* path, unsigned mode, uint32_t max_speed, spi_bit_order_t bit_order, uint8_t bits_per_word, uint8_t extra_flags)`"""

def spi_open_advanced2(spi: intptr, path: str, mode: int, max_speed: int, bit_order: int, bits_per_word: int, extra_flags: int, /) -> int:
    """Wraps `int spi_open_advanced2(spi_t* spi, const char* path, unsigned mode, uint32_t max_speed, spi_bit_order_t bit_order, uint8_t bits_per_word, uint32_t extra_flags)`"""

def spi_transfer(spi: intptr, txbuf: intptr, rxbuf: intptr, len: int, /) -> int:
    """Wraps `int spi_transfer(spi_t* spi, const uint8_t* txbuf, uint8_t* rxbuf, size_t len)`"""

def spi_transfer_advanced(spi: intptr, msgs: intptr, count: int, /) -> int:
    """Wraps `int spi_transfer_advanced(spi_t* spi, const spi_msg_t* msgs, size_t count)`"""

def spi_close(spi: intptr, /) -> int:
    """Wraps `int spi_close(spi_t* spi)`"""

def spi_free(spi: intptr, /) -> None:
    """Wraps `void spi_free(spi_t* spi)`"""

def spi_get_mode(spi: intptr, mode: intptr, /) -> int:
    """Wraps `int spi_get_mode(spi_t* spi, unsigned* mode)`"""

def spi_get_max_speed(spi: intptr, max_speed: intptr, /) -> int:
    """Wraps `int spi_get_max_speed(spi_t* spi, uint32_t* max_speed)`"""

def spi_get_bit_order(spi: intptr, bit_order: intptr, /) -> int:
    """Wraps `int spi_get_bit_order(spi_t* spi, spi_bit_order_t* bit_order)`"""

def spi_get_bits_per_word(spi: intptr, bits_per_word: intptr, /) -> int:
    """Wraps `int spi_get_bits_per_word(spi_t* spi, uint8_t* bits_per_word)`"""

def spi_get_extra_flags(spi: intptr, extra_flags: intptr, /) -> int:
    """Wraps `int spi_get_extra_flags(spi_t* spi, uint8_t* extra_flags)`"""

def spi_get_extra_flags32(spi: intptr, extra_flags: intptr, /) -> int:
    """Wraps `int spi_get_extra_flags32(spi_t* spi, uint32_t* extra_flags)`"""

def spi_set_mode(spi: intptr, mode: int, /) -> int:
    """Wraps `int spi_set_mode(spi_t* spi, unsigned mode)`"""

def spi_set_max_speed(spi: intptr, max_speed: int, /) -> int:
    """Wraps `int spi_set_max_speed(spi_t* spi, uint32_t max_speed)`"""

def spi_set_bit_order(spi: intptr, bit_order: int, /) -> int:
    """Wraps `int spi_set_bit_order(spi_t* spi, spi_bit_order_t bit_order)`"""

def spi_set_bits_per_word(spi: intptr, bits_per_word: int, /) -> int:
    """Wraps `int spi_set_bits_per_word(spi_t* spi, uint8_t bits_per_word)`"""

def spi_set_extra_flags(spi: intptr, extra_flags: int, /) -> int:
    """Wraps `int spi_set_extra_flags(spi_t* spi, uint8_t extra_flags)`"""

def spi_set_extra_flags32(spi: intptr, extra_flags: int, /) -> int:
    """Wraps `int spi_set_extra_flags32(spi_t* spi, uint32_t extra_flags)`"""

def spi_fd(spi: intptr, /) -> int:
    """Wraps `int spi_fd(spi_t* spi)`"""

def spi_tostring(spi: intptr, str: intptr, len: int, /) -> int:
    """Wraps `int spi_tostring(spi_t* spi, char* str, size_t len)`"""

def spi_errno(spi: intptr, /) -> int:
    """Wraps `int spi_errno(spi_t* spi)`"""

def spi_errmsg(spi: intptr, /) -> str:
    """Wraps `const char* spi_errmsg(spi_t* spi)`"""

def periphery_version() -> str:
    """Wraps `const char* periphery_version()`"""

def periphery_version_info() -> intptr:
    """Wraps `const periphery_version_t* periphery_version_info()`"""

# aliases
gpio_direction_t = int
gpio_edge_t = int
gpio_event_clock_t = int
gpio_bias_t = int
gpio_drive_t = int
pwm_polarity_t = int
serial_parity_t = int
spi_bit_order_t = int
# enums
GPIO_ERROR_ARG: int
GPIO_ERROR_OPEN: int
GPIO_ERROR_NOT_FOUND: int
GPIO_ERROR_QUERY: int
GPIO_ERROR_CONFIGURE: int
GPIO_ERROR_UNSUPPORTED: int
GPIO_ERROR_INVALID_OPERATION: int
GPIO_ERROR_IO: int
GPIO_ERROR_CLOSE: int
GPIO_DIR_IN: int
GPIO_DIR_OUT: int
GPIO_DIR_OUT_LOW: int
GPIO_DIR_OUT_HIGH: int
GPIO_EDGE_NONE: int
GPIO_EDGE_RISING: int
GPIO_EDGE_FALLING: int
GPIO_EDGE_BOTH: int
GPIO_EVENT_CLOCK_REALTIME: int
GPIO_EVENT_CLOCK_MONOTONIC: int
GPIO_EVENT_CLOCK_HTE: int
GPIO_BIAS_DEFAULT: int
GPIO_BIAS_PULL_UP: int
GPIO_BIAS_PULL_DOWN: int
GPIO_BIAS_DISABLE: int
GPIO_DRIVE_DEFAULT: int
GPIO_DRIVE_OPEN_DRAIN: int
GPIO_DRIVE_OPEN_SOURCE: int
LED_ERROR_ARG: int
LED_ERROR_OPEN: int
LED_ERROR_QUERY: int
LED_ERROR_IO: int
LED_ERROR_CLOSE: int
MMIO_ERROR_ARG: int
MMIO_ERROR_OPEN: int
MMIO_ERROR_CLOSE: int
PWM_ERROR_ARG: int
PWM_ERROR_OPEN: int
PWM_ERROR_QUERY: int
PWM_ERROR_CONFIGURE: int
PWM_ERROR_CLOSE: int
PWM_POLARITY_NORMAL: int
PWM_POLARITY_INVERSED: int
SERIAL_ERROR_ARG: int
SERIAL_ERROR_OPEN: int
SERIAL_ERROR_QUERY: int
SERIAL_ERROR_CONFIGURE: int
SERIAL_ERROR_IO: int
SERIAL_ERROR_CLOSE: int
PARITY_NONE: int
PARITY_ODD: int
PARITY_EVEN: int
SPI_ERROR_ARG: int
SPI_ERROR_OPEN: int
SPI_ERROR_QUERY: int
SPI_ERROR_CONFIGURE: int
SPI_ERROR_TRANSFER: int
SPI_ERROR_CLOSE: int
SPI_ERROR_UNSUPPORTED: int
MSB_FIRST: int
LSB_FIRST: int


import math

class complex:
    def __init__(self, real, imag=0):
        self._real = float(real)
        self._imag = float(imag)

    @property
    def real(self):
        return self._real
    
    @property
    def imag(self):
        return self._imag

    def conjugate(self):
        return complex(self.real, -self.imag)
    
    def __repr__(self):
        s = ['(', str(self.real)]
        s.append('-' if self.imag < 0 else '+')
        s.append(str(abs(self.imag)))
        s.append('j)')
        return ''.join(s)
    
    def __eq__(self, other):
        if type(other) is complex:
            return self.real == other.real and self.imag == other.imag
        if type(other) in (int, float):
            return self.real == other and self.imag == 0
        return NotImplemented
    
    def __ne__(self, other):
        res = self == other
        if res is NotImplemented:
            return res
        return not res
    
    def __add__(self, other):
        if type(other) is complex:
            return complex(self.real + other.real, self.imag + other.imag)
        if type(other) in (int, float):
            return complex(self.real + other, self.imag)
        return NotImplemented
        
    def __radd__(self, other):
        return self.__add__(other)
    
    def __sub__(self, other):
        if type(other) is complex:
            return complex(self.real - other.real, self.imag - other.imag)
        if type(other) in (int, float):
            return complex(self.real - other, self.imag)
        return NotImplemented
    
    def __rsub__(self, other):
        if type(other) is complex:
            return complex(other.real - self.real, other.imag - self.imag)
        if type(other) in (int, float):
            return complex(other - self.real, -self.imag)
        return NotImplemented
    
    def __mul__(self, other):
        if type(other) is complex:
            return complex(self.real * other.real - self.imag * other.imag,
                           self.real * other.imag + self.imag * other.real)
        if type(other) in (int, float):
            return complex(self.real * other, self.imag * other)
        return NotImplemented
    
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __truediv__(self, other):
        if type(other) is complex:
            denominator = other.real ** 2 + other.imag ** 2
            real_part = (self.real * other.real + self.imag * other.imag) / denominator
            imag_part = (self.imag * other.real - self.real * other.imag) / denominator
            return complex(real_part, imag_part)
        if type(other) in (int, float):
            return complex(self.real / other, self.imag / other)
        return NotImplemented
    
    def __pow__(self, other: int | float):
        if type(other) in (int, float):
            return complex(self.__abs__() ** other * math.cos(other * phase(self)),
                           self.__abs__() ** other * math.sin(other * phase(self)))
        return NotImplemented
    
    def __abs__(self) -> float:
        return math.sqrt(self.real ** 2 + self.imag ** 2)

    def __neg__(self):
        return complex(-self.real, -self.imag)
    
    def __hash__(self):
        return hash((self.real, self.imag))


# Conversions to and from polar coordinates

def phase(z: complex):
    return math.atan2(z.imag, z.real)

def polar(z: complex):
    return z.__abs__(), phase(z)

def rect(r: float, phi: float):
    return r * math.cos(phi) + r * math.sin(phi) * 1j

# Power and logarithmic functions

def exp(z: complex):
    return math.exp(z.real) * rect(1, z.imag)

def log(z: complex, base=2.718281828459045):
    return math.log(z.__abs__(), base) + phase(z) * 1j

def log10(z: complex):
    return log(z, 10)

def sqrt(z: complex):
    return z ** 0.5

# Trigonometric functions

def acos(z: complex):
    return -1j * log(z + sqrt(z * z - 1))

def asin(z: complex):
    return -1j * log(1j * z + sqrt(1 - z * z))

def atan(z: complex):
    return 1j / 2 * log((1 - 1j * z) / (1 + 1j * z))

def cos(z: complex):
    return (exp(1j * z) + exp(-1j * z)) / 2

def sin(z: complex):
    return (exp(1j * z) - exp(-1j * z)) / (2 * 1j)

def tan(z: complex):
    return sin(z) / cos(z)

# Hyperbolic functions

def acosh(z: complex):
    return log(z + sqrt(z * z - 1))

def asinh(z: complex):
    return log(z + sqrt(z * z + 1))

def atanh(z: complex):
    return 1 / 2 * log((1 + z) / (1 - z))

def cosh(z: complex):
    return (exp(z) + exp(-z)) / 2

def sinh(z: complex):
    return (exp(z) - exp(-z)) / 2

def tanh(z: complex):
    return sinh(z) / cosh(z)

# Classification functions

def isfinite(z: complex):
    return math.isfinite(z.real) and math.isfinite(z.imag)

def isinf(z: complex):
    return math.isinf(z.real) or math.isinf(z.imag)

def isnan(z: complex):
    return math.isnan(z.real) or math.isnan(z.imag)

def isclose(a: complex, b: complex):
    return math.isclose(a.real, b.real) and math.isclose(a.imag, b.imag)

# Constants

pi = math.pi
e = math.e
tau = 2 * pi
inf = math.inf
infj = complex(0, inf)
nan = math.nan
nanj = complex(0, nan)



import picoterm
from vmath import rgb

picoterm.enable_full_buffering_mode()

bg = rgb(78, 118, 164)
fg = rgb(200, 200, 0)
text = "hello, \nworld"
text = bg.ansi_bg(text)
text = fg.ansi_fg(text)

def ansi_italic(text: str):
    return f'\x1b[3m{text}\x1b[0m'

text = ansi_italic(text) + '123'
print(text)

cpnts = picoterm.split_ansi_escaped_string(text)

assert cpnts == ['\x1b[3m', '\x1b[38;2;200;200;0m', '\x1b[48;2;78;118;164m', 'hello, ', '\n', 'world', '\x1b[0m', '\x1b[0m', '\x1b[0m', '123']

cpnts_join = ''.join(cpnts)
assert cpnts_join == text

assert picoterm.wcwidth(ord('\n')) == 0
assert picoterm.wcwidth(ord('a')) == 1
assert picoterm.wcwidth(ord('测')) == 2
assert picoterm.wcwidth(ord('👀')) == 2

assert picoterm.wcswidth("hello, 测试a测试👀测\n") == 7 + 1 + 12

text = rgb(12, 34, 56).ansi_fg("hello")
out_list = []
assert picoterm.sscanf(text, "\x1b[38;2;%d;%d;%dm", out_list)
assert out_list == [12, 34, 56]

assert picoterm.sscanf(text, "\x1b[38;2;%d;%d;%dmhello", out_list)
assert out_list == [12, 34, 56]

assert picoterm.sscanf(text, "\x1b[38;2;%d;%d;%d", out_list)
assert not picoterm.sscanf(text, "\x1b[38;2;%d;%d;%dm???", out_list)
assert not picoterm.sscanf(text, "\x1b[77;2;%d;%d;%dm", out_list)


# test super:
class TestSuperBase():
    def __init__(self):
        self.base_attr = 1
        
    def base_method(self):
        return self.base_attr
    
    def error(self):
        raise RuntimeError('未能拦截错误')
    

class TestSuperChild1(TestSuperBase):
    def __init__(self):
        super(TestSuperChild1, self).__init__()
    
    def child_method(self):
        return super(TestSuperChild1, self).base_method()
    
    def error_handling(self):
        try:
            super(TestSuperChild1, self).error()
        except RuntimeError:
            pass

class TestSuperChild2(TestSuperBase):
    pass


test_base = TestSuperBase()
# 测试属性
assert test_base.base_attr == 1
# 测试方法
assert test_base.base_method() == 1

test_child1 = TestSuperChild1()
# 测试继承的属性
assert test_child1.base_attr == 1
# 测试继承的方法
assert test_child1.base_method() == 1
# 测试子类添加的方法
assert test_child1.child_method() == 1
# 测试子类的错误拦截
test_child1.error_handling()

test_child2 = TestSuperChild2()
# 测试继承的属性
assert test_child2.base_attr == 1
# 测试继承的方法
assert test_child2.base_method() == 1


class TestSuperNoBaseMethod(TestSuperBase):
    def __init__(self):
        super(TestSuperNoBaseMethod, self).append(1)

class TestSuperNoParent():
    def method(self):
        super(TestSuperNoParent, self).method()

try:
    t = TestSuperNoParent().method()
    print('未能拦截错误2')
    exit(2)
except AttributeError:
    pass

try:
    t = TestSuperNoBaseMethod()
    print('未能拦截错误3')
    exit(3)
except AttributeError:
    pass

class B():
    pass

class C():
    def method(self):
        super(C, self).method()

class D():
    def method(self):
        super(B, self).method()

try:
    c = C()
    c.method()
    print('未能拦截错误4')
    exit(4)
except AttributeError:
    pass

try:
    d = D()
    d.method()
    print('未能拦截错误5')
    exit(5)
except TypeError:
    pass

# test hash:
# 测试整数类型的输入
assert type(hash(0)) is int
assert type(hash(123)) is int
assert type(hash(-456)) is int

# 测试字符串类型的输入
assert type(hash("hello")) is int

# 测试浮点数类型的输入
assert type(hash(3.14)) is int 
assert type(hash(-2.71828)) is int

# 测试边界情况
# assert type(hash(None)) is int 
assert hash(True) == 1
assert hash(False) == 0

# 测试元组
assert type(hash((4, 5, 6, (1234,1122), 2.3983, 'abcd'))) is int

# 测试自定义类和对象的输入
class A():
    pass

a = A()

assert type(hash(A)) is int
assert type(hash(a)) is int

# 测试函数的输入
def f():
    pass
assert type(hash(a)) is int

# 测试不可哈希对象
try:
    hash({1:1})
    print('未能拦截错误6')
    exit(6)
except TypeError:
    pass

try:
    hash([1])
    print('未能拦截错误7')
    exit(7)
except TypeError:
    pass

# test chr
actual = []
for i in range(128):
    actual.append(f'{i} {chr(i)}')
expected = ['0 \x00', '1 \x01', '2 \x02', '3 \x03', '4 \x04', '5 \x05', '6 \x06', '7 \x07', '8 \x08', '9 \t', '10 \n', '11 \x0b', '12 \x0c', '13 \r', '14 \x0e', '15 \x0f', '16 \x10', '17 \x11', '18 \x12', '19 \x13', '20 \x14', '21 \x15', '22 \x16', '23 \x17', '24 \x18', '25 \x19', '26 \x1a', '27 \x1b', '28 \x1c', '29 \x1d', '30 \x1e', '31 \x1f', '32  ', '33 !', '34 "', '35 #', '36 $', '37 %', '38 &', "39 '", '40 (', '41 )', '42 *', '43 +', '44 ,', '45 -', '46 .', '47 /', '48 0', '49 1', '50 2', '51 3', '52 4', '53 5', '54 6', '55 7', '56 8', '57 9', '58 :', '59 ;', '60 <', '61 =', '62 >', '63 ?', '64 @', '65 A', '66 B', '67 C', '68 D', '69 E', '70 F', '71 G', '72 H', '73 I', '74 J', '75 K', '76 L', '77 M', '78 N', '79 O', '80 P', '81 Q', '82 R', '83 S', '84 T', '85 U', '86 V', '87 W', '88 X', '89 Y', '90 Z', '91 [', '92 \\', '93 ]', '94 ^', '95 _', '96 `', '97 a', '98 b', '99 c', '100 d', '101 e', '102 f', '103 g', '104 h', '105 i', '106 j', '107 k', '108 l', '109 m', '110 n', '111 o', '112 p', '113 q', '114 r', '115 s', '116 t', '117 u', '118 v', '119 w', '120 x', '121 y', '122 z', '123 {', '124 |', '125 }', '126 ~', '127 \x7f']
assert len(actual) == len(expected)
for i in range(len(actual)):
    assert (actual[i] == expected[i]), (actual[i], expected[i])

# assert type(bin(1234)) is str

# test __repr__:
class A():
    def __init__(self):
        self.attr = 0

repr(A())

try:
    range(1,2,3,4)
    print('未能拦截错误8, 在测试 range')
    exit(8)
except TypeError:
    pass

# /************ int ************/
try:
    int('asad')
    print('未能拦截错误9, 在测试 int')
    exit(9)
except ValueError:
    pass

try:
    int(123, 16)
    print('未能拦截错误10, 在测试 int')
    exit(10)
except TypeError:
    pass

assert type(10//11) is int

assert type(11%2) is int

try:
    float('asad')
    print('未能拦截错误11, 在测试 float')
    exit(11)
except ValueError:
    pass

try:
    float([])
    print('未能拦截错误12, 在测试 float')
    exit(12)
except TypeError:
    pass

# /************ str ************/
# test str.__rmul__:
assert type(12 * '12') is str

# 未完全测试准确性-----------------------------------------------
# test str.index:
assert type('25363546'.index('63')) is int
try:
    '25363546'.index('err')
    print('未能拦截错误13, 在测试 str.index')
    exit(13)
except ValueError as e:
    assert str(e) == "substring not found"


# 未完全测试准确性-----------------------------------------------
# test str.find:
assert '25363546'.find('63') == 3
assert '25363546'.find('err') == -1


# /************ list ************/
try:
    list(1,2)
    print('未能拦截错误14, 在测试 list')
    exit(14)
except TypeError:
    pass

# 未完全测试准确性----------------------------------------------
# test list.index:
assert type([1,2,3,4,5].index(4)) is int
try:
    [1,2,3,4,5].index(6)
    print('未能拦截错误15, 在测试 list.index')
    exit(15)
except ValueError as e:
    assert str(e) == "list.index(x): x not in list"



# 未完全测试准确性----------------------------------------------
# test list.remove:
try:
    [1,2,3,4,5].remove(6)
    print('未能拦截错误16, 在测试 list.remove')
    exit(16)
except ValueError as e:
    assert str(e) == "list.remove(x): x not in list"


# 未完全测试准确性----------------------------------------------
# test list.pop:
try:
    [1,2,3,4,5].pop(1,2,3,4)
    print('未能拦截错误17, 在测试 list.pop')
    exit(17)
except TypeError:
    pass



# 未完全测试准确性-----------------------------------------------
# test list.__rmul__:
assert type(12 * [12]) is list


# /************ tuple ************/
# test tuple:
try:
    tuple(1,2)
    print('未能拦截错误18, 在测试 tuple')
    exit(18)
except TypeError:
    pass

assert [1,2,2,3,3,3].count(3) == 3
assert [1,2,2,3,3,3].count(0) == 0
assert 3 in (1, 3, 4)
assert 5 not in (1, 3, 4)

assert repr(True) == 'True'
assert repr(False) == 'False'

assert True & True == True

assert True | False == True

assert (True ^ True) == False

assert (True == True) == True

assert type(hash(bytes([0x41, 0x42, 0x43]))) is int


# 未完全测试准确性-----------------------------------------------
# test bytes.__repr__:
assert type(repr(bytes([0x41, 0x42, 0x43]))) is str


# /************ slice ************/
assert type(slice(0.1, 0.2, 0.3)) is slice



# test reduce

from functools import reduce, partial

# test reduce
assert reduce(lambda x, y: x + y, [1, 2, 3, 4, 5]) == 15
assert reduce(lambda x, y: x * y, [1, 2, 3, 4, 5]) == 120
assert reduce(lambda x, y: x * y, [1, 2, 3, 4, 5], 10) == 1200

# test partial
def add(a, b):
    return a + b

add_1 = partial(add, 1)

assert add_1(2) == 3
assert add_1(3) == 4

def sub(a, b=1):
    return a - b

sub_10 = partial(sub, b=10)
assert sub_10(20) == 10
assert sub_10(30) == 20

# test lru_cache
from functools import lru_cache
miss_keys = []

@lru_cache(maxsize=3)
def test_f(x: int):
    miss_keys.append(x)
    return x

assert test_f(1) == 1 and miss_keys == [1]
# [1]
assert test_f(2) == 2 and miss_keys == [1, 2]
# [1, 2]
assert test_f(3) == 3 and miss_keys == [1, 2, 3]
# [1, 2, 3]
assert test_f(1) == 1 and miss_keys == [1, 2, 3]
# [2, 3, 1]
assert test_f(2) == 2 and miss_keys == [1, 2, 3]
# [3, 1, 2]
assert test_f(4) == 4 and miss_keys == [1, 2, 3, 4]
# [1, 2, 4]
assert test_f(3) == 3 and miss_keys == [1, 2, 3, 4, 3]
# [2, 4, 3]
assert test_f(2) == 2 and miss_keys == [1, 2, 3, 4, 3]
# [4, 3, 2]
assert test_f(5) == 5 and miss_keys == [1, 2, 3, 4, 3, 5]
# [3, 2, 5]
assert test_f(3) == 3 and miss_keys == [1, 2, 3, 4, 3, 5]
# [2, 5, 3]
assert test_f(1) == 1 and miss_keys == [1, 2, 3, 4, 3, 5, 1]
# [5, 3, 1]



import os

def get_loc(path):
    loc = 0
    with open(path, "rt", encoding='utf-8') as f:
        loc += len(f.readlines())
    path = path.replace('\\', '/')
    print(f'{path}: {loc}')
    return loc

def get_loc_for_dir(path):
    loc = 0
    loc_ex = 0
    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith('.h') or file.endswith('.c') or file.endswith('.h'):
                _i = get_loc(os.path.join(root, file))
                if file.startswith('_'):
                    loc_ex += _i
                else:
                    loc += _i
    return f'{path}: {loc} (+{loc_ex})'


print(get_loc_for_dir('include/pocketpy'))
print()
print(get_loc_for_dir('src'))



import os

class MyClass:
    value = os.environ['TEST_RELOAD_VALUE']

    def some_func(self):
        return self.value
    
    @staticmethod
    def get_xy():
        g = globals()
        return g.get('x', 0), g.get('y', 0)


if os.environ['SET_X'] == '1':
    x = 1
elif os.environ['SET_Y'] == '1':
    y = 1


x = 0
if x==0: x=1
assert x==1

def f(x, y): return x+y
assert f(1,2)==3

c = 1
if c==0: x=1
elif c==1: x=2
else: x=3
assert x==2

def f1(x): return x+1
assert f1(1)==2

# class A: pass
# class B: pass
# a = A()
# assert type(a) is A


from unicodedata import east_asian_width

# full width
assert east_asian_width("Ａ") == "F"
# half width
assert east_asian_width("ｻ") == "H"
# narrow
assert east_asian_width("a") == "Na"
# wide
assert east_asian_width("测") == "W"
assert east_asian_width("🥕") == "W"
assert east_asian_width("。") == "W"
# ambiguous
assert east_asian_width("°") == "A"
# neutral
assert east_asian_width("\n") == "N"




from pkpy import configmacros

if configmacros['PK_ENABLE_THREADS'] == 1:
    from pkpy import ComputeThread
else:
    print('threads is not enabled, skipping test...')
    exit()

import time

thread_1 = ComputeThread(1)
thread_2 = ComputeThread(2)

for t in [thread_1, thread_2]:
    t.exec('''
def func(a):
    from pkpy import currentvm
    print("Hello from thread", currentvm(), "a =", a)
    for i in range(500000):
        if i % 100000 == 0:
            print(i, "from thread", currentvm())
    return a
                   
x = 123
''')
assert thread_1.eval('x') == 123

# thread_1.wait_for_done()
# thread_2.wait_for_done()

thread_1.submit_call('func', [1, 2, 3])
thread_2.submit_call('func', [4, 5, 6])

while not thread_1.is_done or not thread_2.is_done:
    print("Waiting for threads to finish...")
    time.sleep(1)

print("Thread 1 last return value:", thread_1.last_retval())
print("Thread 2 last return value:", thread_2.last_retval())




def enable_full_buffering_mode() -> None:
    """Enable full buffering mode for ASCII drawings (32KB)."""

def split_ansi_escaped_string(s: str) -> list[str]:
    """Perform split on ANSI escaped string."""

def wcwidth(c: int) -> int: ...
def wcswidth(s: str) -> int: ...

def sscanf(s: str, fmt: str, out_list: list) -> bool: ...


def f():
    raise IndexError

try:
    f()
    exit(1)
except IndexError:
    pass

k = KeyError('foo')
assert str(k) == "'foo'"

try:
    assert False
    exit(1)
except AssertionError:
    pass

try:
    raise 1
except TypeError:
    pass

try:
    for i in range(5):
        raise KeyError(i)
    exit(1)
except KeyError:
    pass

x = 0
for i in range(5):
    try:
        for j in range(5):
            while True:
                raise KeyError(i)
                x += 3
    except KeyError:
        x += i
assert x == 10

class A:
    def __getitem__(self, i):
        raise KeyError(i)

try:
    a = A()
    b = a[1]
    exit(1)
except KeyError:
    pass

try:
    a = {'1': 3, 4: None}
    x = a[1]
    exit(1)
except:
    pass
assert True

def f():
    try:
        raise KeyError('foo')
    except IndexError:   # will fail to catch
        exit(1)
    except:
        pass
    assert True

f()

def f1():
    try:
        assert 1 + 2 == 3
        try:
            a = {1: 2, 3: 4}
            x = a[0]
        except RuntimeError:
            exit(1)
    except IndexError:
        exit(1)
    exit(1)

try:
    f1()
    exit(1)
except KeyError:
    pass


assert True, "Msg"
try:
    assert False, "Msg"
    exit(1)
except AssertionError:
    pass

def f(a: list):
    try:
        raise ValueError
        exit(1)
    except:
        pass
    a[0] = 1
a = [0]
f(a)
assert a == [1]

try:
    a = [][3]
except IndexError as e:
    # assert str(e) == '3 not in [0, 0)'
    assert repr(e).startswith('IndexError(')

try:
    a = {}[2]
except IndexError as e:
    exit(1)
except Exception as e:
    assert type(e) is KeyError
    assert str(e) == '2'
    assert repr(e).startswith('KeyError(')
except:
    exit(1)


class MyException(Exception):
    pass

class MyException2(MyException):
    pass

try:
    raise MyException2
except MyException as e:
    ok = True
except Exception:
    exit(1)
assert ok

ok = False
try:
    eval('1+')
except SyntaxError as e:
    assert type(e) is SyntaxError
    ok = True
assert ok

# nested try
def g():
    try:
        raise KeyError
    except KeyError:
        pass

try:
    raise IndexError
except IndexError:
    g()

a = []
for i in range(10):
    a.append(i)
    try:
        try:
            if i % 2 == 0:
                raise KeyError(i)
            else:
                raise IndexError(i)
        except KeyError as e:
            assert i % 2 == 0
            assert e.args[0] == i
            raise
        except IndexError as e:
            assert i % 2 == 1
            assert e.args[0] == i
            raise
    except Exception as e:
        assert e.args[0] == i
assert a == list(range(10))

# inner exc
x = 0
try:
    try:
        [][1]
    except:
        raise KeyError
except KeyError:
    x = 5
assert x == 5

a = []
for i in range(6):
    try:
        [][1]
    except IndexError:
        if i == 2:
            continue
        else:
            a.append(i)
assert a == [0, 1, 3, 4, 5]

try:
    result = 10 / 0
    exit(1)
except (ZeroDivisionError, TypeError) as e:
    if type(e) != ZeroDivisionError:
        exit(1)

try:
    try:
        result = 10 / 0
        exit(1)
    except (ZeroDivisionError, 1) as e:
        exit(1)
    except (TypeError) as e:
        exit(1)
except Exception as e:
    if type(e) != TypeError:
        exit(1)

"""
# finally, only
def finally_only():
    try:
        raise KeyError
    finally:
        x = 1

try:
    finally_only()
    exit(1)
except KeyError:
    pass

def finally_only_2():
    x = 0
    try:
        pass
    finally:
        x = 1
    return x
    
assert finally_only_2() == 1

# finally, no exception
def finally_no_exception():
    ok = False
    try:
        pass
    except KeyError:
        exit(1)
    finally:
        ok = True
    return ok

assert finally_no_exception()

# finally, match
def finally_match():
    ok = False
    try:
        raise KeyError
    except KeyError:
        pass
    finally:
        ok = True
    return ok

assert finally_match()

# finally, no match
ok = False
def finally_no_match():
    global ok
    try:
        raise KeyError
    except IndexError:
        exit(1)
    finally:
        ok = True

ok_2 = False
try:
    finally_no_match()
except KeyError:
    assert ok
    ok_2 = True

assert ok_2

# finally, return
def finally_return():
    try:
        raise KeyError
    finally:
        return 1
    
assert finally_return() == 1
"""



class Writer:
    def __init__(self) -> None:
        self.buffer = []
        self.indent_level = 0

    def indent(self):
        self.indent_level += 1

    def dedent(self):
        self.indent_level -= 1

    def write(self, line: str):
        self.buffer.append('    ' * self.indent_level + line)

    def __str__(self) -> str:
        return '\n'.join(self.buffer)


path = []

class A:
    def __init__(self, x):
        self.x = x
        self.path = []

    def __enter__(self):
        path.append('enter')
        return self.x
    
    def __exit__(self, *args):
        path.append('exit')
    

with A(123):
    assert path == ['enter']
assert path == ['enter', 'exit']

path.clear()

with A(123) as a:
    assert path == ['enter']
    assert a == 123
    path.append('in')
assert path == ['enter', 'in', 'exit']

path.clear()





class cache:
    def __init__(self, f):
        self.f = f
        self.cache = {}

    def __call__(self, *args):
        if args not in self.cache:
            self.cache[args] = self.f(*args)
        return self.cache[args]
    
class lru_cache:
    def __init__(self, maxsize=128):
        self.maxsize = maxsize
        self.cache = {}

    def __call__(self, f):
        def wrapped(*args):
            if args in self.cache:
                res = self.cache.pop(args)
                self.cache[args] = res
                return res
            
            res = f(*args)
            if len(self.cache) >= self.maxsize:
                first_key = next(iter(self.cache))
                self.cache.pop(first_key)
            self.cache[args] = res
            return res
        return wrapped
    
def reduce(function, sequence, initial=...):
    it = iter(sequence)
    if initial is ...:
        try:
            value = next(it)
        except StopIteration:
            raise TypeError("reduce() of empty sequence with no initial value")
    else:
        value = initial
    for element in it:
        value = function(value, element)
    return value

class partial:
    def __init__(self, f, *args, **kwargs):
        self.f = f
        if not callable(f):
            raise TypeError("the first argument must be callable")
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        kwargs.update(self.kwargs)
        return self.f(*self.args, *args, **kwargs)




from .writer import Writer
from .converters import get_converter
from .schema import Function

from keyword import iskeyword

def sanitize_name(name: str):
    if iskeyword(name):
        return name + '_'
    return name

def gen_function(w: Writer, pyi_w: Writer, function: Function):
    name = function.name
    w.write(f'static bool cfunc__{name}(int argc, py_Ref argv) {{')
    w.indent()

    w.write(f'PY_CHECK_ARGC({len(function.params)});')

    args_cvt = [get_converter(arg.type) for arg in function.params]
    ret_cvt = get_converter(function.ret_type)

    for i in range(len(args_cvt)):
        w.write(f'{args_cvt[i].T} _{i};')
        args_cvt[i].py2c(w, f'_{i}', f'py_arg({i})')

    call_args = ', '.join([f'_{i}' for i in range(len(args_cvt))])

    # gen retval
    if function.ret_type == 'void':
        w.write(f'{name}({call_args});')
        w.write(f'py_newnone(py_retval());')
    else:
        w.write(f'{ret_cvt.T} res = {name}({call_args});')
        ret_cvt.c2py(w, 'py_retval()', 'res')

    w.write('return true;')

    w.dedent()
    w.write('}')

    # pyi
    py_args = []
    arg_names = [sanitize_name(arg.name) for arg in function.params]
    for i in range(len(args_cvt)):
        py_args.append(f'{arg_names[i]}: {args_cvt[i].py_T}')

    if len(py_args) > 0:
        py_args.append('/')
    pyi_w.write(f'def {name}({", ".join(py_args)}) -> {ret_cvt.py_T}:')
    if function.desc:
        pyi_w.write(f'    """Wraps `{function.signature()}`\n\n    {function.desc}"""')
    else:
        pyi_w.write(f'    """Wraps `{function.signature()}`"""')
    pyi_w.write('')


def _get_annotations(cls: type):
    inherits = []
    while cls is not object:
        inherits.append(cls)
        cls = cls.__base__
    inherits.reverse()
    res = {}
    for cls in inherits:
        res.update(cls.__annotations__)
    return res.keys()

def _wrapped__init__(self, *args, **kwargs):
    cls = type(self)
    cls_d = cls.__dict__
    fields = _get_annotations(cls)
    i = 0   # index into args
    for field in fields:
        if field in kwargs:
            setattr(self, field, kwargs.pop(field))
        else:
            if i < len(args):
                setattr(self, field, args[i])
                i += 1
            elif field in cls_d:    # has default value
                setattr(self, field, cls_d[field])
            else:
                raise TypeError(f"{cls.__name__} missing required argument {field!r}")
    if len(args) > i:
        raise TypeError(f"{cls.__name__} takes {len(fields)} positional arguments but {len(args)} were given")
    if len(kwargs) > 0:
        raise TypeError(f"{cls.__name__} got an unexpected keyword argument {next(iter(kwargs))!r}")

def _wrapped__repr__(self):
    fields = _get_annotations(type(self))
    obj_d = self.__dict__
    args: list = [f"{field}={obj_d[field]!r}" for field in fields]
    return f"{type(self).__name__}({', '.join(args)})"

def _wrapped__eq__(self, other):
    if type(self) is not type(other):
        return False
    fields = _get_annotations(type(self))
    for field in fields:
        if getattr(self, field) != getattr(other, field):
            return False
    return True

def _wrapped__ne__(self, other):
    return not self.__eq__(other)

def dataclass(cls: type):
    assert type(cls) is type
    cls_d = cls.__dict__
    if '__init__' not in cls_d:
        cls.__init__ = _wrapped__init__
    if '__repr__' not in cls_d:
        cls.__repr__ = _wrapped__repr__
    if '__eq__' not in cls_d:
        cls.__eq__ = _wrapped__eq__
    if '__ne__' not in cls_d:
        cls.__ne__ = _wrapped__ne__
    fields = _get_annotations(cls)
    has_default = False
    for field in fields:
        if field in cls_d:
            has_default = True
        else:
            if has_default:
                raise TypeError(f"non-default argument {field!r} follows default argument")
    return cls

def asdict(obj) -> dict:
    fields = _get_annotations(type(obj))
    obj_d = obj.__dict__
    return {field: obj_d[field] for field in fields}


def isgeneratorfunction(obj) -> bool: ...

def is_user_defined_type(t: type) -> bool:
    """Check if a type is user-defined.
    
    This means the type was created by executing python `class` statement."""


from pycparser import c_ast

class Pointer:
    def __init__(self, base: str, level: int):
        # super().__init__(f'{base}' + '*' * level)
        self.base = base
        self.level = level

class NamedFields:
    def __init__(self, name: str, fields: dict[str, str] | None, typedef_name: str | None = None):
        self.name = name
        self.fields = fields
        self.typedef_name = typedef_name

    def is_opaque(self):
        return self.fields is None
    
    def __repr__(self):
        cls = type(self).__name__
        return f"{cls}('{self.name}', {self.fields!r})"
    
class Struct(NamedFields): pass
class Union(NamedFields): pass
    
class Enum:
    def __init__(self, name: str, typedef_name: str | None = None):
        self.name = name
        self.typedef_name = typedef_name
        self.values = [] # type: list[str]

    def __repr__(self):
        return f"Enum('{self.name}', values={self.values!r})"

class Function:
    def __init__(self, name: str, ret: str):
        self.name = name
        self.args = [] # type: list[tuple[str, str]]
        self.ret = ret

    def __repr__(self):
        return f"Function('{self.name}', args={self.args!r}, ret={self.ret!r})"


import os
import subprocess

def get_all_files(root: str):
    for path, _, files in os.walk(root):
        for file in files:
            fullpath = os.path.join(path, file)
            # ignore some files
            if fullpath.startswith('include/pybind11'):
                continue
            if file.startswith('_'):
                continue
            if not file.endswith('.c') and not file.endswith('.h') and not file.endswith('.hpp'):
                continue
            yield fullpath

if __name__ == '__main__':
    files = []
    files.extend(get_all_files('include'))
    files.extend(get_all_files('src'))
    files.extend(get_all_files('src2'))
    subprocess.run(['clang-format', '-i'] + files, check=True)



def is_prime(x):
  if x<2:
    return False
  for i in range(2,x):
    if x%i == 0:
      return False
  return True

def test(n):
  k = 0
  for i in range(n):
    if is_prime(i):
      k += 1
  return k

# from dis import dis
# dis(test)
# dis(is_prime)

assert test(10000) == 1229


import math

# ============================================================
# Helper function for floating point comparison
# ============================================================

def isclose(a, b):
    return abs(a-b) < 0.000001

# ============================================================
# Constants: pi, e, inf, nan
# ============================================================

assert isclose(math.pi, 3.141592653589793)
assert isclose(math.e, 2.718281828459045)
assert math.isinf(math.inf)
assert math.isnan(math.nan)

# ============================================================
# Rounding functions: ceil, floor, trunc, fabs
# ============================================================

# ceil - round up to nearest integer
assert math.ceil(1.2) == 2
assert math.ceil(-1.2) == -1
assert math.ceil(2.0) == 2
assert math.ceil(0) == 0

# floor - round down to nearest integer
assert math.floor(1.2) == 1
assert math.floor(-1.2) == -2
assert math.floor(2.0) == 2
assert math.floor(0) == 0

# trunc - truncate towards zero
assert math.trunc(1.7) == 1
assert math.trunc(-1.7) == -1
assert math.trunc(2.0) == 2
assert math.trunc(0) == 0

# fabs - absolute value (returns float)
assert isclose(math.fabs(-1.2), 1.2)
assert isclose(math.fabs(1.2), 1.2)
assert isclose(math.fabs(0), 0.0)
assert isclose(math.fabs(-0.0), 0.0)

# ============================================================
# Classification functions: isnan, isinf, isfinite, isclose
# ============================================================

# isnan - check if value is NaN
assert math.isnan(float('nan'))
assert math.isnan(math.nan)
assert not math.isnan(1.0)
assert not math.isnan(math.inf)

# isinf - check if value is infinity
assert math.isinf(float('inf'))
assert math.isinf(float('-inf'))
assert math.isinf(math.inf)
assert not math.isinf(1.0)
assert not math.isinf(math.nan)

# isfinite - check if value is finite (not inf or nan)
assert math.isfinite(1.0)
assert math.isfinite(0.0)
assert math.isfinite(-1.0)
assert not math.isfinite(math.inf)
assert not math.isfinite(-math.inf)
assert not math.isfinite(math.nan)

# isclose - check if two values are close
assert math.isclose(1.0, 1.0)
assert math.isclose(1.0, 1.0 + 1e-10)
assert not math.isclose(1.0, 1.1)

# ============================================================
# Power and logarithmic functions: exp, log, log2, log10, pow, sqrt
# ============================================================

# exp - e raised to the power x
assert isclose(math.exp(0), 1.0)
assert isclose(math.exp(1), math.e)
assert isclose(math.exp(2), math.e * math.e)

# log - natural logarithm (base e) or logarithm with custom base
assert isclose(math.log(1), 0.0)
assert isclose(math.log(math.e), 1.0)
assert isclose(math.log(10), 2.302585092994046)
assert isclose(math.log(100, 10), 2.0)  # log base 10
assert isclose(math.log(8, 2), 3.0)     # log base 2

# log2 - logarithm base 2
assert isclose(math.log2(1), 0.0)
assert isclose(math.log2(2), 1.0)
assert isclose(math.log2(8), 3.0)
assert isclose(math.log2(10), 3.321928094887362)

# log10 - logarithm base 10
assert isclose(math.log10(1), 0.0)
assert isclose(math.log10(10), 1.0)
assert isclose(math.log10(100), 2.0)
assert isclose(math.log10(1000), 3.0)

# pow - x raised to the power y
assert isclose(math.pow(2, 3), 8.0)
assert isclose(math.pow(2, 0), 1.0)
assert isclose(math.pow(2, -1), 0.5)
assert isclose(math.pow(4, 0.5), 2.0)
assert isclose(math.pow(0, 0), 1.0)

# sqrt - square root
assert isclose(math.sqrt(4), 2.0)
assert isclose(math.sqrt(2), 1.4142135623730951)
assert isclose(math.sqrt(0), 0.0)
assert isclose(math.sqrt(1), 1.0)

# ============================================================
# Trigonometric functions: sin, cos, tan, asin, acos, atan, atan2
# ============================================================

# sin - sine
assert isclose(math.sin(0), 0.0)
assert isclose(math.sin(math.pi / 2), 1.0)
assert isclose(math.sin(math.pi), 0.0)
assert isclose(math.sin(-math.pi / 2), -1.0)
assert isclose(math.sin(-math.pi / 4), -0.7071067811865476)

# cos - cosine
assert isclose(math.cos(0), 1.0)
assert isclose(math.cos(math.pi / 2), 0.0)
assert isclose(math.cos(math.pi), -1.0)
assert isclose(math.cos(-math.pi), -1.0)

# tan - tangent
assert isclose(math.tan(0), 0.0)
assert isclose(math.tan(math.pi / 4), 1.0)
assert isclose(math.tan(-math.pi / 4), -1.0)

# asin - arc sine (inverse sine)
assert isclose(math.asin(0), 0.0)
assert isclose(math.asin(1), math.pi / 2)
assert isclose(math.asin(-1), -math.pi / 2)
assert isclose(math.asin(0.5), math.pi / 6)

# acos - arc cosine (inverse cosine)
assert isclose(math.acos(1), 0.0)
assert isclose(math.acos(0), math.pi / 2)
assert isclose(math.acos(-1), math.pi)
assert isclose(math.acos(0.5), math.pi / 3)

# atan - arc tangent (inverse tangent)
assert isclose(math.atan(0), 0.0)
assert isclose(math.atan(1), math.pi / 4)
assert isclose(math.atan(-1), -math.pi / 4)

# atan2 - arc tangent of y/x (handles quadrants correctly)
assert isclose(math.atan2(0, 1), 0.0)
assert isclose(math.atan2(1, 0), math.pi / 2)
assert isclose(math.atan2(0, -1), math.pi)
assert isclose(math.atan2(-1, 0), -math.pi / 2)
assert isclose(math.atan2(1, 1), math.pi / 4)
assert isclose(math.atan2(-1, -1), -3 * math.pi / 4)

# ============================================================
# Angle conversion functions: degrees, radians
# ============================================================

# degrees - convert radians to degrees
assert isclose(math.degrees(0), 0.0)
assert isclose(math.degrees(math.pi), 180.0)
assert isclose(math.degrees(math.pi / 2), 90.0)
assert isclose(math.degrees(2 * math.pi), 360.0)

# radians - convert degrees to radians
assert isclose(math.radians(0), 0.0)
assert isclose(math.radians(180), math.pi)
assert isclose(math.radians(90), math.pi / 2)
assert isclose(math.radians(360), 2 * math.pi)

# ============================================================
# Arithmetic functions: fmod, modf, copysign, fsum, gcd, factorial
# ============================================================

# fmod - floating point modulo
assert isclose(math.fmod(5.0, 3.0), 2.0)
assert isclose(math.fmod(-5.0, 3.0), -2.0)
assert isclose(math.fmod(5.0, -3.0), 2.0)
assert isclose(math.fmod(-5.0, -3.0), -2.0)
assert isclose(math.fmod(4.0, 3.0), 1.0)
assert isclose(math.fmod(-4.0, 3.0), -1.0)
assert isclose(math.fmod(-2.0, 3.0), -2.0)
assert isclose(math.fmod(2.0, 3.0), 2.0)

# modf - split into fractional and integer parts
frac, integer = math.modf(1.5)
assert isclose(frac, 0.5)
assert isclose(integer, 1.0)

frac, integer = math.modf(-1.5)
assert isclose(frac, -0.5)
assert isclose(integer, -1.0)

frac, integer = math.modf(2.0)
assert isclose(frac, 0.0)
assert isclose(integer, 2.0)

frac, integer = math.modf(0.0)
assert isclose(frac, 0.0)
assert isclose(integer, 0.0)

# copysign - return x with the sign of y
assert isclose(math.copysign(1.0, -1.0), -1.0)
assert isclose(math.copysign(-1.0, 1.0), 1.0)
assert isclose(math.copysign(1.0, 1.0), 1.0)
assert isclose(math.copysign(-1.0, -1.0), -1.0)
assert isclose(math.copysign(5.0, -0.0), -5.0)
assert isclose(math.copysign(0.0, -1.0), 0.0)  # sign of zero may vary

# fsum - accurate floating point sum
assert math.fsum([0.1] * 10) == 1.0
assert math.fsum([]) == 0.0
assert math.fsum([1.0, 2.0, 3.0]) == 6.0
assert math.fsum([0.1, 0.2, 0.3]) == 0.6

# gcd - greatest common divisor
assert math.gcd(10, 5) == 5
assert math.gcd(10, 6) == 2
assert math.gcd(10, 7) == 1
assert math.gcd(10, 10) == 10
assert math.gcd(-10, 10) == 10
assert math.gcd(10, -10) == 10
assert math.gcd(-10, -10) == 10
assert math.gcd(0, 5) == 5
assert math.gcd(5, 0) == 5
assert math.gcd(0, 0) == 0

# factorial - n!
assert math.factorial(0) == 1
assert math.factorial(1) == 1
assert math.factorial(2) == 2
assert math.factorial(3) == 6
assert math.factorial(4) == 24
assert math.factorial(5) == 120
assert math.factorial(10) == 3628800

# ============================================================
# Special value tests
# ============================================================

# Test NaN generation from invalid operations
a = -0.1
a = a ** a
assert math.isnan(a)
assert not math.isinf(a)

# Test infinity
assert math.isinf(float("inf"))
assert math.isinf(float("-inf"))
assert not math.isnan(float("inf"))


# Test str.__mod__ (old-style % formatting)

# Test %s - string formatting
assert "hello %s" % "world" == "hello world"
assert "%s" % 123 == "123"
assert "%s %s" % ("a", "b") == "a b"
assert "name: %s, age: %s" % ("Alice", 30) == "name: Alice, age: 30"

# Test %d - integer formatting
assert "%d" % 42 == "42"
assert "%d" % -123 == "-123"
assert "%d" % 0 == "0"
assert "count: %d" % 100 == "count: 100"

# Test %i - integer formatting (same as %d)
assert "%i" % 42 == "42"
assert "%i" % -123 == "-123"
assert "value: %i" % 999 == "value: 999"

# Test %f - float formatting
assert "%f" % 3.14 == "3.140000"
assert "%f" % -2.5 == "-2.500000"
assert "%f" % 0.0 == "0.000000"
assert "%f" % 42 == "42.000000"  # int to float

# Test %r - repr formatting
assert "%r" % "hello" == "'hello'"
assert "%r" % 123 == "123"
assert "%r" % [1, 2, 3] == "[1, 2, 3]"

# Test %% - literal percent
assert "%%" % () == "%"
assert "100%%" % () == "100%"
assert "%%s" % () == "%s"
assert "%d%%" % 50 == "50%"

# Test combined format specifiers
assert "%s is %d years old" % ("Bob", 25) == "Bob is 25 years old"
assert "%s: %f" % ("pi", 3.14159) == "pi: 3.141590"
assert "%d + %d = %d" % (1, 2, 3) == "1 + 2 = 3"
assert "Hello %s! You have %d messages." % ("User", 5) == "Hello User! You have 5 messages."

# Test single value (not tuple)
assert "value: %s" % "test" == "value: test"
assert "number: %d" % 42 == "number: 42"

# Test empty string
assert "" % () == ""

# Test no format specifiers
assert "hello world" % () == "hello world"
